import unittest

from odml import Property, Section, Document
from odml.doc import BaseDocument
from odml.section import BaseSection


class TestSection(unittest.TestCase):

    def setUp(self):
        pass

    def test_simple_attributes(self):
        sec_name = "sec name"
        sec_type = "sec type"
        sec_def = "an odml test section"
        sec_ref = "from over there"

        sec = Section(name=sec_name, type=sec_type, definition=sec_def, reference=sec_ref)

        self.assertEqual(sec.name, sec_name)
        self.assertEqual(sec.type, sec_type)
        self.assertEqual(sec.definition, sec_def)
        self.assertEqual(sec.reference, sec_ref)

        # Test setting attributes
        sec.name = "%s_edit" % sec_name
        self.assertEqual(sec.name, "%s_edit" % sec_name)
        sec.type = "%s_edit" % sec_type
        self.assertEqual(sec.type, "%s_edit" % sec_type)
        sec.definition = "%s_edit" % sec_def
        self.assertEqual(sec.definition, "%s_edit" % sec_def)
        sec.reference = "%s_edit" % sec_ref
        self.assertEqual(sec.reference, "%s_edit" % sec_ref)

        # Test setting attributes to None when '' is passed.
        sec.reference = ""
        self.assertIsNone(sec.reference)
        sec.definition = ""
        self.assertIsNone(sec.definition)

    def test_name(self):
        # Test id is used when name is not provided
        s = Section()
        self.assertIsNotNone(s.name)
        self.assertEqual(s.name, s.id)

        # Test name is properly set on init
        name = "rumpelstilzchen"
        s = Section(name)
        self.assertEqual(s.name, name)

        name = "rumpelstilzchen"
        s = Section(name=name)
        self.assertEqual(s.name, name)

        # Test name can be properly set on single and connected Sections
        sec = Section()
        self.assertNotEqual(sec.name, "sec")
        sec.name = "sec"
        self.assertEqual(sec.name, "sec")

        subsec_a = Section(parent=sec)
        self.assertNotEqual(subsec_a.name, "subsec_a")
        subsec_a.name = "subsec_a"
        self.assertEqual(subsec_a.name, "subsec_a")

        # Test subsection name can be changed with siblings
        subsec_b = Section(name="subsec_b", parent=sec)
        self.assertEqual(subsec_b.name, "subsec_b")
        subsec_b.name = "subsec"
        self.assertEqual(subsec_b.name, "subsec")

        # Test subsection name set will fail on existing sibling with same name
        with self.assertRaises(KeyError):
            subsec_b.name = "subsec_a"
        self.assertEqual(subsec_b.name, "subsec")

        # Test section name set will fail on existing same name document sibling
        doc = Document()
        sec_a = Section(name="a", parent=doc)
        sec_b = Section(name="b", parent=doc)
        with self.assertRaises(KeyError):
            sec_b.name = "a"

    def test_parent(self):
        s = Section("Section")
        self.assertIsNone(s.parent)
        s.parent = None
        self.assertIsNone(s.parent)
        s.parent = Document()
        self.assertIsInstance(s.parent, BaseDocument)
        self.assertIsInstance(s.parent._sections[0], BaseSection)

        """ Test if child is removed from _sections of a parent after assigning
            a new parent to the child """
        p = s.parent
        s.parent = Section("S")
        self.assertEqual(len(p._sections), 0)

        s = Section("section_doc", parent=Document())
        self.assertIsInstance(s.parent, BaseDocument)
        self.assertEqual(len(s.parent._sections), 1)
        p = s.parent
        s.parent = None
        self.assertEqual(len(p._sections), 0)
        self.assertEqual(s.parent, None)

        s = Section("section_sec", parent=Section("S"))
        self.assertIsInstance(s.parent, BaseSection)
        self.assertEqual(len(s.parent._sections), 1)
        p = s.parent
        s.parent = None
        self.assertEqual(len(p._sections), 0)
        self.assertEqual(s.parent, None)

        with self.assertRaises(ValueError):
            Section("section_property", parent=Property("P"))

    def test_path(self):
        sec = Section(name="center")
        self.assertEqual(sec.get_path(), "/")

        subsec = Section(name="leaf", parent=sec)
        self.assertEqual(subsec.get_path(), "/leaf")

        doc = Document()
        sec.parent = doc
        self.assertEqual(sec.get_path(), "/center")
        self.assertEqual(subsec.get_path(), "/center/leaf")

        top = Section(name="top", parent=doc)
        sec.parent = top
        self.assertEqual(sec.get_path(), "/top/center")
        self.assertEqual(subsec.get_path(), "/top/center/leaf")

        subsec.parent = None
        self.assertEqual(subsec.get_path(), "/")

    def test_children(self):
        sec = Section(name="sec")

        # Test set sections
        subsec = Section(name="subsec", parent=sec)
        newsec = Section(name="newsec")

        self.assertEqual(subsec.parent, sec)
        self.assertEqual(sec.sections[0], subsec)
        self.assertEqual(len(sec.sections), 1)
        self.assertIsNone(newsec.parent)

        sec.sections[0] = newsec
        self.assertEqual(newsec.parent, sec)
        self.assertEqual(sec.sections[0], newsec)
        self.assertEqual(len(sec.sections), 1)
        self.assertIsNone(subsec.parent)

        # Test parent cleanup
        root = Section(name="root")
        sec.parent = root
        subsec.parent = newsec

        self.assertEqual(len(newsec.sections), 1)
        self.assertEqual(newsec.sections[0], subsec)
        self.assertEqual(subsec.parent, newsec)
        self.assertEqual(len(root.sections), 1)
        self.assertEqual(root.sections[0], sec)

        subsec.parent = root
        self.assertEqual(len(newsec.sections), 0)
        self.assertEqual(subsec.parent, root)
        self.assertEqual(len(root.sections), 2)
        self.assertEqual(root.sections[1], subsec)

        # Test set section fails
        with self.assertRaises(ValueError):
            sec.sections[0] = Document()
        with self.assertRaises(ValueError):
            sec.sections[0] = Property("fail")
        with self.assertRaises(ValueError):
            sec.sections[0] = "subsec"

        # Test set properties
        prop = Property(name="prop", parent=sec)
        newprop = Property(name="newprop")

        self.assertEqual(prop.parent, sec)
        self.assertEqual(sec.properties[0], prop)
        self.assertEqual(len(sec.properties), 1)
        self.assertIsNone(newprop.parent)

        sec.properties[0] = newprop
        self.assertEqual(newprop.parent, sec)
        self.assertEqual(sec.properties[0], newprop)
        self.assertEqual(len(sec.properties), 1)
        self.assertIsNone(prop.parent)

        # Test set property fails
        with self.assertRaises(ValueError):
            sec.properties[0] = Document()
        with self.assertRaises(ValueError):
            sec.properties[0] = newsec
        with self.assertRaises(ValueError):
            sec.properties[0] = "prop"

    def test_id(self):
        s = Section(name="S")
        self.assertIsNotNone(s.id)

        s = Section("S", oid="79b613eb-a256-46bf-84f6-207df465b8f7")
        self.assertEqual(s.id, "79b613eb-a256-46bf-84f6-207df465b8f7")

        s = Section("S", oid="id")
        self.assertNotEqual(s.id, "id")

        # Make sure id cannot be reset programmatically.
        with self.assertRaises(AttributeError):
            s.id = "someId"

    def test_new_id(self):
        sec = Section(name="sec")
        old_id = sec.id

        # Test assign new generated id.
        sec.new_id()
        self.assertNotEqual(old_id, sec.id)

        # Test assign new custom id.
        old_id = sec.id
        sec.new_id("79b613eb-a256-46bf-84f6-207df465b8f7")
        self.assertNotEqual(old_id, sec.id)
        self.assertEqual("79b613eb-a256-46bf-84f6-207df465b8f7", sec.id)

        # Test invalid custom id exception.
        with self.assertRaises(ValueError):
            sec.new_id("crash and burn")

    def test_clone(self):
        # Check parent removal in clone.
        psec = Section(name="parent")
        sec = Section(name="original", parent=psec)
        clone_sec = sec.clone()
        self.assertEqual(sec.parent, psec)
        self.assertIsNone(clone_sec.parent)

        # Check new id in clone.
        sec = Section(name="original")
        clone_sec = sec.clone()
        self.assertEqual(sec, clone_sec)
        self.assertNotEqual(sec.id, clone_sec.id)

        # Check child Sections and Properties are cloned and have new ids.
        Section(name="sec_one", parent=sec)
        Section(name="sec_two", parent=sec)
        Property(name="prop_one", parent=sec)
        Property(name="prop_two", parent=sec)

        clone_sec = sec.clone()

        # Check sections
        self.assertListEqual(sec.sections, clone_sec.sections)
        self.assertEqual(sec.sections["sec_one"], clone_sec.sections["sec_one"])
        self.assertNotEqual(sec.sections["sec_one"].id, clone_sec.sections["sec_one"].id)

        # Check properties
        self.assertListEqual(sec.properties, clone_sec.properties)
        self.assertEqual(sec.properties["prop_one"], clone_sec.properties["prop_one"])
        self.assertNotEqual(sec.properties["prop_one"].id,
                            clone_sec.properties["prop_one"].id)

        # Check child Sections and Properties are not cloned.
        clone_sec = sec.clone(children=False)

        self.assertListEqual(clone_sec.sections, [])
        self.assertListEqual(clone_sec.properties, [])

    def test_reorder(self):
        # Test reorder of document sections
        doc = Document()
        sec_one = Section(name="sec_one", parent=doc)
        sec_two = Section(name="sec_two", parent=doc)
        sec_three = Section(name="sec_three", parent=doc)

        self.assertEqual(doc.sections[0].name, sec_one.name)
        self.assertEqual(doc.sections[2].name, sec_three.name)
        sec_three.reorder(0)

        self.assertEqual(doc.sections[0].name, sec_three.name)
        self.assertEqual(doc.sections[2].name, sec_two.name)

        # Test reorder of document sections
        sec = Section(name="main")
        sec_one = Section(name="sec_one", parent=sec)
        sec_two = Section(name="sec_two", parent=sec)
        sec_three = Section(name="sec_three", parent=sec)

        self.assertEqual(sec.sections[0].name, sec_one.name)
        self.assertEqual(sec.sections[2].name, sec_three.name)
        sec_three.reorder(0)

        self.assertEqual(sec.sections[0].name, sec_three.name)
        self.assertEqual(sec.sections[2].name, sec_two.name)

        # Test Exception on unconnected section
        with self.assertRaises(ValueError):
            sec.reorder(0)

    def test_append(self):
        main = Section(name="main")
        self.assertListEqual(main.sections, [])
        self.assertListEqual(main.properties, [])

        # Test append Section
        sec = Section(name="sec1")
        main.append(sec)
        self.assertEqual(len(main.sections), 1)
        self.assertEqual(sec.parent, main)

        # Test fail on Section list or tuple append
        with self.assertRaises(ValueError):
            main.append([Section(name="sec2"), Section(name="sec3")])
        with self.assertRaises(ValueError):
            main.append((Section(name="sec2"), Section(name="sec3")))
        self.assertEqual(len(main.sections), 1)

        # Test append Property
        prop = Property(name="prop")
        main.append(prop)
        self.assertEqual(len(main.properties), 1)
        self.assertEqual(prop.parent, main)

        # Test fail on Property list or tuple append
        with self.assertRaises(ValueError):
            main.append([Property(name="prop2"), Property(name="prop3")])
        with self.assertRaises(ValueError):
            main.append((Property(name="prop2"), Property(name="prop3")))
        self.assertEqual(len(main.properties), 1)

        # Test fail on unsupported value
        with self.assertRaises(ValueError):
            main.append(Document())
        with self.assertRaises(ValueError):
            main.append("Section")

        # Test fail on same name entities
        with self.assertRaises(KeyError):
            main.append(Section(name="sec1"))
        self.assertEqual(len(main.sections), 1)

        with self.assertRaises(KeyError):
            main.append(Property(name="prop"))
        self.assertEqual(len(main.properties), 1)

    def test_extend(self):
        sec = Section(name="main")
        self.assertListEqual(sec.sections, [])
        self.assertListEqual(sec.properties, [])

        # Test extend with Section list
        sec.extend([Section(name="sec1"), Section(name="sec2")])
        self.assertEqual(len(sec), 2)
        self.assertEqual(len(sec.sections), 2)
        self.assertEqual(sec.sections[0].name, "sec1")

        # Test extend with Property list
        sec.extend((Property(name="prop1"), Property(name="prop2")))
        self.assertEqual(len(sec), 4)
        self.assertEqual(len(sec.properties), 2)
        self.assertEqual(sec.properties[0].name, "prop1")

        # Test extend with mixed list
        sec.extend([Section(name="sec3"), Property(name="prop3")])
        self.assertEqual(len(sec), 6)
        self.assertEqual(len(sec.sections), 3)
        self.assertEqual(len(sec.properties), 3)

        # Test fail on non iterable
        with self.assertRaises(TypeError):
            sec.extend(1)

        # Test fail on non Section/Property list entry
        with self.assertRaises(ValueError):
            sec.extend([Property(name="prop4"), 5])

        # Test fail on same name entities
        with self.assertRaises(KeyError):
            sec.extend([Property(name="new"), Property(name="prop3")])
        self.assertEqual(len(sec.properties), 3)

        with self.assertRaises(KeyError):
            sec.extend([Section(name="new"), Section(name="sec3")])
        self.assertEqual(len(sec.sections), 3)

    def test_remove(self):
        sec = Section(name="remsec")

        ssec_one = Section(name="subsec_one", parent=sec)
        ssec_two = Section(name="subsec_two", parent=sec)
        self.assertEqual(len(sec.sections), 2)
        self.assertIsNotNone(ssec_one.parent)

        sec.remove(ssec_one)
        self.assertEqual(len(sec.sections), 1)
        self.assertEqual(sec.sections[0].name, ssec_two.name)
        self.assertIsNone(ssec_one.parent)

        with self.assertRaises(ValueError):
            sec.remove(ssec_one)
        self.assertEqual(len(sec.sections), 1)

        prop_one = Property(name="prop_one", parent=sec)
        prop_two = Property(name="prop_two", parent=sec)
        self.assertEqual(len(sec.properties), 2)
        self.assertIsNotNone(prop_one.parent)

        sec.remove(prop_one)
        self.assertEqual(len(sec.properties), 1)
        self.assertEqual(sec.properties[0].name, prop_two.name)
        self.assertIsNone(prop_one.parent)

        with self.assertRaises(ValueError):
            sec.remove(prop_one)
        self.assertEqual(len(sec.properties), 1)

        with self.assertRaises(ValueError):
            sec.remove("prop_two")

    def test_insert(self):
        sec = Section(name="root")

        _ = Section(name="subsec_one", parent=sec)
        _ = Section(name="subsec_two", parent=sec)
        subsec = Section(name="subsec_three")

        self.assertNotEqual(sec.sections[1].name, subsec.name)
        sec.insert(1, subsec)
        self.assertEqual(len(sec.sections), 3)
        self.assertEqual(sec.sections[1].name, subsec.name)
        self.assertEqual(subsec.parent.name, sec.name)

        _ = Property(name="prop_one", parent=sec)
        _ = Property(name="prop_two", parent=sec)
        prop = Property(name="prop_three")

        self.assertNotEqual(sec.properties[1].name, prop.name)
        sec.insert(1, prop)
        self.assertEqual(len(sec.properties), 3)
        self.assertEqual(sec.properties[1].name, prop.name)
        self.assertEqual(prop.parent.name, sec.name)

        # Test invalid object
        with self.assertRaises(ValueError):
            sec.insert(1, Document())
        with self.assertRaises(ValueError):
            sec.insert(1, "some info")
        self.assertEqual(len(sec), 6)

        # Test same name entries
        with self.assertRaises(ValueError):
            sec.insert(0, subsec)
        with self.assertRaises(ValueError):
            sec.insert(0, prop)
        self.assertEqual(len(sec), 6)

    def test_contains(self):
        sec = Section(name="root")

        subsec = Section(name="subsec", type="test")
        prop = Property(name="prop")

        # Test not contains on empty child-lists.
        self.assertIsNone(sec.contains(subsec))
        self.assertIsNone(sec.contains(prop))

        # Test contains of Section and Property
        subsec.parent = sec
        simisec = Section(name="subsec", type="test")
        self.assertEqual(sec.contains(simisec).name, subsec.name)

        prop.parent = sec
        simiprop = Property(name="prop")
        self.assertEqual(sec.contains(simiprop).name, prop.name)

        # Test not contains on mismatching Section name/type and Property name
        self.assertIsNone(sec.contains(Section(name="subsec", type="typetest")))
        self.assertIsNone(sec.contains(Section(name="typesec", type="test")))
        self.assertIsNone(sec.contains(Property(name="prop_two")))

        # Test fail on non-Section/Property objects
        with self.assertRaises(ValueError):
            sec.contains(Document())

        with self.assertRaises(ValueError):
            sec.contains("some info")

    def test_merge_check(self):
        # -- Root level Section checks

        # Test empty Section check
        source = Section(name="source")
        destination = Section(name="destination")

        destination.merge_check(source, True)

        # Test definition check
        source = Section(name="source", definition="def")
        destination = Section(name="destination", definition="def")

        destination.merge_check(source, True)
        source.definition = "other def"
        with self.assertRaises(ValueError):
            destination.merge_check(source, True)

        # Test reference check
        source = Section(name="source", reference="ref")
        destination = Section(name="destination", reference="ref")

        destination.merge_check(source, True)
        source.reference = "other ref"
        with self.assertRaises(ValueError):
            destination.merge_check(source, True)

        # -- First child level Section checks
        source = Section(name="source")
        destination = Section(name="destination")

        s_sec_one = Section(name="lvl", type="one",
                            reference="ref", definition="def", parent=source)
        s_sec_two = Section(name="unrelated", type="one",
                            reference="one", definition="one", parent=source)

        d_sec_one = Section(name="lvl", type="one",
                            reference="ref", definition="def", parent=destination)
        d_sec_two = Section(name="unrelated", type="two",
                            reference="two", definition="two", parent=destination)

        # Test Section child level definition check
        destination.merge_check(source, True)
        s_sec_one.definition = "other def"
        with self.assertRaises(ValueError):
            destination.merge_check(source, True)

        # Test Section child level reference check
        s_sec_one.definition = "def"
        s_sec_one.reference = "other ref"
        with self.assertRaises(ValueError):
            destination.merge_check(source, True)

        # -- Second child level Section checks
        source = Section(name="source")
        destination = Section(name="destination")

        s_sec_one = Section(name="lvl", type="one",
                            reference="ref", definition="def", parent=source)
        s_subsec_one = Section(name="lvl", type="two",
                               reference="ref2", definition="def2", parent=s_sec_one)
        s_sec_two = Section(name="unrelated", type="one",
                            reference="one", definition="one", parent=source)
        s_subsec_two = Section(name="lvl", type="two",
                               reference="none1", definition="none1", parent=s_sec_two)

        d_sec_one = Section(name="lvl", type="one",
                            reference="ref", definition="def", parent=destination)
        d_subsec_one = Section(name="lvl", type="two",
                               reference="ref2", definition="def2", parent=d_sec_one)
        d_sec_two = Section(name="unrelated", type="two",
                            reference="two", definition="two", parent=destination)
        d_subsec_two = Section(name="lvl", type="two",
                               reference="none2", definition="none2", parent=d_sec_two)

        # Test Section 2nd child level definition check
        # Check no definition/reference ValueError between s_subsec_two and d_subsec_one
        # since their parents will not be merged.
        destination.merge_check(source, True)

        # Raise a definition ValueError between s_subsec_one and d_subsec_one
        # since their parents will be merged.
        s_subsec_one.definition = "other def"
        with self.assertRaises(ValueError):
            destination.merge_check(source, True)

        # Test Section 2nd child level reference check
        s_subsec_one.definition = "def2"

        # Raise a reference ValueError between s_subsec_one and d_subsec_one
        # since their parents will be merged.
        s_subsec_one.reference = "other ref"
        with self.assertRaises(ValueError):
            destination.merge_check(source, True)

        # -- Root level Property checks
        # All Property checks will only test unit failure in the Section merge context.
        # Other failures are covered by the specific Property merge check tests.
        source = Section(name="source")
        destination = Section(name="destination")

        s_prop = Property(name="prop", parent=source)
        d_prop = Property(name="prop", parent=destination)

        destination.merge_check(source, True)
        s_prop.unit = "Hz"
        d_prop.unit = "s"
        with self.assertRaises(ValueError):
            destination.merge_check(source, True)

        # -- First child level Property checks
        source = Section(name="source")
        destination = Section(name="destination")

        s_prop_one = Property(name="lvl one", unit="Hz", parent=source)
        s_prop_two = Property(name="unrelated one", unit="one", parent=source)

        d_prop_one = Property(name="lvl one", unit="Hz", parent=destination)
        d_prop_two = Property(name="unrelated two", unit="two", parent=destination)

        # Test Property child level check
        destination.merge_check(source, True)

        # Test raise ValueError between s_prop_one and d_prop_one
        s_prop_one.unit = "other unit"
        with self.assertRaises(ValueError):
            destination.merge_check(source, True)

        # -- Second child level Property checks
        source = Section(name="source")
        destination = Section(name="destination")

        s_sec_one = Section(name="lvl", type="one", parent=source)
        s_subprop_one = Property(name="lvl one", unit="Hz", parent=s_sec_one)

        s_sec_two = Section(name="unrelated", type="one", parent=source)
        s_subprop_two = Property(name="unrelated one", unit="one", parent=s_sec_two)

        d_sec_one = Section(name="lvl", type="one", parent=destination)
        d_subprop_one = Property(name="lvl one", unit="Hz", parent=d_sec_one)

        d_sec_two = Section(name="unrelated", type="two", parent=destination)
        d_subprop_two = Property(name="unrelated one", unit="two", parent=d_sec_two)

        # Test Property 2nd child level definition check
        # Check no unit ValueError between s_subprop_two and d_subprop_one
        # since their parents will not be merged.
        destination.merge_check(source, True)

        # Raise a unit ValueError between s_subprop_one and d_subprop_one
        # since their parents will be merged.
        s_subprop_one.unit = "other unit"
        with self.assertRaises(ValueError):
            destination.merge_check(source, True)

    def test_merge(self):
        # -- Root level Section merge tests
        source = Section(name="source", definition="def", reference="ref")
        destination = Section(name="destination")

        destination.merge(source)
        self.assertEqual(destination.definition, source.definition)
        self.assertEqual(destination.reference, source.reference)

        # -- First child level Section merge tests
        s_sec_one = Section(name="lvl", type="one", definition="def", parent=source)
        s_sec_two = Section(name="other", type="one", parent=source)
        d_sec_one = Section(name="lvl", type="one", parent=destination)

        self.assertEqual(len(destination), 1)
        self.assertIsNone(destination.sections["lvl"].definition)
        self.assertIsNone(destination.sections["lvl"].reference)

        destination.merge(source)
        self.assertEqual(len(destination), 2)
        self.assertEqual(destination.sections["lvl"].definition, s_sec_one.definition)
        self.assertEqual(destination.sections["lvl"].reference, s_sec_one.reference)
        self.assertEqual(destination.sections["other"], s_sec_two)

        # -- Root level Property merge tests
        source = Section(name="source")
        destination = Section(name="destination")

        s_prop_one = Property(name="prop_one", unit="Hz", parent=source)
        s_prop_two = Property(name="prop_two", parent=source)
        d_prop_one = Property(name="prop_one", parent=destination)

        self.assertEqual(len(destination.properties), 1)
        self.assertIsNone(destination.properties["prop_one"].unit)

        destination.merge(source)
        self.assertEqual(len(destination.properties), 2)
        self.assertEqual(destination.properties["prop_one"].unit, s_prop_one.unit)
        self.assertEqual(destination.properties["prop_two"], s_prop_two)

        # -- First child level Property merge tests
        source = Section(name="source")
        destination = Section(name="destination")

        s_sec_one = Section(name="lvl", type="one", definition="def", parent=source)
        s_prop_one = Property(name="prop_one", unit="Hz", parent=s_sec_one)
        s_prop_two = Property(name="prop_two", parent=s_sec_one)

        d_sec_one = Section(name="lvl", type="one", parent=destination)
        d_prop_one = Property(name="prop_one", parent=d_sec_one)

        self.assertEqual(len(destination.properties), 0)
        self.assertEqual(len(destination.sections["lvl"].properties), 1)
        self.assertIsNone(destination.sections["lvl"].properties["prop_one"].unit)

        destination.merge(source)
        self.assertEqual(len(destination.properties), 0)
        self.assertEqual(len(destination.sections["lvl"].properties), 2)
        self.assertEqual(destination.sections["lvl"].properties["prop_one"].unit,
                         s_prop_one.unit)
        self.assertEqual(destination.sections["lvl"].properties["prop_two"],
                         s_prop_two)

        # -- Test nothing merged on second child level ValueError
        source = Section(name="source", definition="def", reference="ref")
        destination = Section(name="destination")

        s_sec_one = Section(name="lvl", type="one", definition="def", parent=source)
        s_sec_two = Section(name="other", type="one", parent=source)
        d_sec_one = Section(name="lvl", type="one", parent=destination)

        s_subprop_one = Property(name="prop", value=[1, 2, 3], parent=s_sec_one)
        d_subprop_one = Property(name="prop", value=["four", "five"], parent=d_sec_one)

        self.assertEqual(len(destination.sections), 1)
        self.assertEqual(len(destination.sections["lvl"].properties), 1)
        self.assertIsNone(destination.definition)
        self.assertIsNone(destination.sections["lvl"].definition)
        self.assertEqual(destination.sections["lvl"].properties[0].value,
                         d_subprop_one.value)

        with self.assertRaises(ValueError):
            destination.merge(source)

        self.assertEqual(len(destination.sections), 1)
        self.assertEqual(len(destination.sections["lvl"].properties), 1)
        self.assertIsNone(destination.definition)
        self.assertIsNone(destination.sections["lvl"].definition)
        self.assertEqual(destination.sections["lvl"].properties[0].value,
                         d_subprop_one.value)

    def test_comparison(self):
        sec_name = "sec name"
        sec_type = "sec type"
        sec_def = "an odml test section"
        sec_ref = "from over there"

        sec_a = Section(name=sec_name, type=sec_type,
                        definition=sec_def, reference=sec_ref)
        sec_b = Section(name=sec_name, type=sec_type,
                        definition=sec_def, reference=sec_ref)

        self.assertEqual(sec_a, sec_b)

        sec_b.name = "newSecName"
        self.assertNotEqual(sec_a, sec_b)

        sec_b.name = sec_name

        # Test section equality with subsections

        # Test equality with subsection of different entities
        # with same content and same children order
        subsec_aa = Section(name="subsecA", type=sec_type,
                            definition=sec_def, reference=sec_ref)
        subsec_ab = Section(name="subsecB", type=sec_type,
                            definition=sec_def, reference=sec_ref)
        subsec_ba = Section(name="subsecA", type=sec_type,
                            definition=sec_def, reference=sec_ref)
        subsec_bb = Section(name="subsecB", type=sec_type,
                            definition=sec_def, reference=sec_ref)

        sec_a.extend([subsec_aa, subsec_ab])
        sec_b.extend([subsec_ba, subsec_bb])

        self.assertEqual(sec_a, sec_b)
        self.assertEqual(sec_a.sections, sec_b.sections)

        sec_b.sections[0].name = "newSubsecA"
        self.assertNotEqual(sec_a, sec_b)
        self.assertNotEqual(sec_a.sections, sec_b.sections)

        sec_b.sections[0].name = "subsecA"

        # Test inequality with different number of children
        sec_b.remove(sec_b.sections[1])
        self.assertNotEqual(sec_a, sec_b)
        self.assertNotEqual(sec_a.sections, sec_b.sections)

        # Test equality with subsection of different entities
        # with same content and different children order
        sec_b.remove(sec_b.sections[0])
        sec_b.extend([subsec_bb, subsec_ba])

        self.assertEqual(sec_a, sec_b)
        self.assertEqual(sec_a.sections, sec_b.sections)

        sec_b.sections[0].name = "newSubsecB"
        self.assertNotEqual(sec_a, sec_b)
        self.assertNotEqual(sec_a.sections, sec_b.sections)

        sec_b.sections[0].name = "subsecB"

        # Test section equality with properties

        # Test equality with properties of different entities
        # with same content and same children order
        prop_aa = Property(name="propA", definition="propDef")
        prop_ab = Property(name="propB", definition="propDef")
        prop_ba = Property(name="propA", definition="propDef")
        prop_bb = Property(name="propB", definition="propDef")

        sec_a.extend([prop_aa, prop_ab])
        sec_b.extend([prop_ba, prop_bb])

        self.assertEqual(sec_a, sec_b)
        self.assertEqual(sec_a.properties, sec_b.properties)

        sec_b.properties[0].name = "newPropA"
        self.assertNotEqual(sec_a, sec_b)
        self.assertNotEqual(sec_a.properties, sec_b.properties)

        sec_b.properties[0].name = "propA"

        # Test inequality with different number of children
        sec_b.remove(sec_b.properties[1])
        self.assertNotEqual(sec_a, sec_b)
        self.assertNotEqual(sec_a.properties, sec_b.properties)

        # Test equality with properties of different entities
        # with same content and different children order
        sec_b.remove(sec_b.properties[0])
        sec_b.extend([prop_bb, prop_ba])

        self.assertEqual(sec_a, sec_b)
        self.assertEqual(sec_a.properties, sec_b.properties)

        sec_b.properties[0].name = "newPropB"
        self.assertNotEqual(sec_a, sec_b)
        self.assertNotEqual(sec_a.properties, sec_b.properties)

    def test_link(self):
        pass

    def test_include(self):
        pass

    def test_repository(self):
        pass

    def test_unmerge(self):
        pass



import datetime
import os
import unittest
from os.path import dirname, abspath

import yaml
from rdflib import URIRef, Literal
from rdflib.namespace import XSD, RDF

import odml
import odml.format as format
from odml.tools.rdf_converter import RDFWriter
from test.test_samplefile import SampleFileCreator
from test.test_samplefile import parse

odmlns = format.Format.namespace()


class TestRDFWriter(unittest.TestCase):

    def setUp(self):
        doc = SampleFileCreator().create_document()
        doc1 = SampleFileCreator().create_document()
        self.doc = doc
        self.doc1 = doc1

    def test_convert_to_rdf(self):
        w = RDFWriter([self.doc, self.doc1])
        w.convert_to_rdf()
        doc_subjects = w.g.subjects(predicate=RDF.type, object=URIRef(odmlns.Document))
        self.assertEqual(len(list(doc_subjects)), 2)

    def test_adding_doc_to_the_hub(self):
        w = RDFWriter([self.doc])
        w.convert_to_rdf()
        hub_hasDocument = w.g.objects(subject=w.hub_root, predicate=odmlns.hasDocument)
        self.assertEqual(len(list(hub_hasDocument)), 1)

    def test_adding_repository(self):
        w = RDFWriter([self.doc])
        w.convert_to_rdf()
        self.assertEqual(len(list(w.g.objects(subject=w.hub_root, predicate=odmlns.hasTerminology))), 0)
        self.assertEqual(len(list(w.g.objects(subject=URIRef(odmlns + w.docs[0].id), predicate=odmlns.hasTerminology))), 0)

        url = "terminology_url"
        self.doc.repository = url
        w = RDFWriter([self.doc])
        w.convert_to_rdf()
        self.assertEqual(len(list(w.g.subjects(predicate=RDF.type, object=URIRef(url)))), 1)
        self.assertEqual(len(list(w.g.objects(subject=w.hub_root, predicate=odmlns.hasTerminology))), 1)
        self.assertEqual(len(list(w.g.objects(subject=URIRef(odmlns + w.docs[0].id), predicate=odmlns.hasTerminology))), 1)

    def test_adding_sections(self):
        doc = odml.Document()
        w = RDFWriter([doc])
        w.convert_to_rdf()
        self.assertEqual(len(list(w.g.subject_objects(predicate=odmlns.hasSection))), 0)

        w = RDFWriter([self.doc])
        w.convert_to_rdf()
        self.assertEqual(len(list(w.g.subject_objects(predicate=odmlns.hasSection))), 9)

        w = RDFWriter([self.doc, self.doc1])
        w.convert_to_rdf()
        self.assertEqual(len(list(w.g.subject_objects(predicate=odmlns.hasSection))), 18)

    def test_adding_properties(self):
        doc = parse("""
            s1[t1]
            - s11[t1]
            s2[t2]
            """)
        w = RDFWriter([doc])
        w.convert_to_rdf()
        self.assertEqual(len(list(w.g.subject_objects(predicate=odmlns.hasProperty))), 0)

        w = RDFWriter([self.doc])
        w.convert_to_rdf()
        self.assertEqual(len(list(w.g.subject_objects(predicate=odmlns.hasProperty))), 12)

        w = RDFWriter([self.doc, self.doc1])
        w.convert_to_rdf()
        self.assertEqual(len(list(w.g.subject_objects(predicate=odmlns.hasProperty))), 24)

    def test_adding_values(self):
        doc = parse("""
            s1[t1]
            """)

        w = RDFWriter([doc])
        w.convert_to_rdf()
        self.assertEqual(len(list(w.g.subject_objects(predicate=RDF.li))), 0)
        self.assertEqual(len(list(
            w.g.subject_objects(predicate=URIRef("%s_1" % str(RDF))))), 0)

        doc = parse("""
            s1[t1]
            - p1
            """)

        w = RDFWriter([doc])
        w.convert_to_rdf()
        self.assertEqual(len(list(w.g.subjects(predicate=RDF.li,
                                               object=Literal("val")))), 0)
        self.assertEqual(len(list(w.g.subjects(predicate=URIRef("%s_1" % str(RDF)),
                                               object=Literal("val")))), 1)

        doc.sections[0].properties[0].append("val2")
        w = RDFWriter([doc])
        w.convert_to_rdf()
        self.assertEqual(len(list(w.g.subject_objects(predicate=RDF.li))), 0)
        self.assertEqual(len(list(w.g.subjects(predicate=RDF.li, object=Literal("val")))), 0)
        self.assertEqual(len(list(w.g.subjects(predicate=RDF.li, object=Literal("val2")))), 0)

        self.assertEqual(len(list(w.g.subjects(predicate=URIRef("%s_1" % str(RDF)),
                                               object=Literal("val")))), 1)
        self.assertEqual(len(list(w.g.subjects(predicate=URIRef("%s_2" % str(RDF)),
                                               object=Literal("val2")))), 1)

        doc = parse("""
             s1[t1]
             - p1
             s2[t2]
             - p1
             - p2
             """)

        w = RDFWriter([doc])
        w.convert_to_rdf()
        self.assertEqual(len(list(w.g.subjects(predicate=RDF.li, object=Literal("val")))), 0)
        self.assertEqual(len(list(w.g.subjects(predicate=URIRef("%s_1" % str(RDF)),
                                               object=Literal("val")))), 3)

    def test_section_subclass(self):
        p = os.path.join(dirname(dirname(abspath(__file__))), 'doc', 'section_subclasses.yaml')
        with open(p, "r") as f:
            subclass = yaml.load(f)

        doc = odml.Document()
        subclass_key = next(iter(subclass))
        s = odml.Section("S", type=subclass_key)
        doc.append(s)
        w = RDFWriter(doc)
        w.convert_to_rdf()
        self.assertEqual(len(list(w.g.subjects(predicate=RDF.type, object=URIRef(odmlns[subclass[subclass_key]])))), 1)
        self.assertEqual(len(list(w.g.subjects(predicate=RDF.type, object=URIRef(odmlns.Section)))), 0)

    def test_adding_other_entities_properties(self):
        doc = parse("""
            s1[t1]
            - p1
            """)

        version = "v1"
        date = datetime.date(1979, 10, 12)
        author = "nice person"
        s_def = "comment"
        s_ref = "reference"
        p_unit = "u1"
        p_name = "p1"
        p_def = "p comment"
        p_uncertainty = "un"
        p_dtype = "string"
        p_value_origin = "value"
        p_ref = "p_ref"

        doc.version = version
        doc.date = date
        doc.author = author
        doc.sections[0].definition = s_def
        doc.sections[0].reference = s_ref
        doc.sections[0].properties[0].name = p_name
        doc.sections[0].properties[0].unit = p_unit
        doc.sections[0].properties[0].definition = p_def
        doc.sections[0].properties[0].uncertainty = p_uncertainty
        doc.sections[0].properties[0].dtype = p_dtype
        doc.sections[0].properties[0].value_origin = p_value_origin
        doc.sections[0].properties[0].reference = p_ref

        w = RDFWriter([doc])
        w.convert_to_rdf()
        self.assertEqual(len(list(w.g.subjects(predicate=odmlns.hasDocVersion, object=Literal(version)))), 1)
        self.assertEqual(len(list(w.g.subjects(predicate=odmlns.hasDate, object=Literal(date, datatype=XSD.date)))), 1)
        self.assertEqual(len(list(w.g.subjects(predicate=odmlns.hasAuthor, object=Literal(author)))), 1)

        self.assertEqual(len(list(w.g.subjects(predicate=odmlns.hasName, object=Literal("s1")))), 1)
        self.assertEqual(len(list(w.g.subjects(predicate=odmlns.hasType, object=Literal("t1")))), 1)
        self.assertEqual(len(list(w.g.subjects(predicate=odmlns.hasDefinition, object=Literal(s_def)))), 1)
        self.assertEqual(len(list(w.g.subjects(predicate=odmlns.hasReference, object=Literal(s_ref)))), 1)

        self.assertEqual(len(list(w.g.subjects(predicate=odmlns.hasName, object=Literal(p_name)))), 1)
        self.assertEqual(len(list(w.g.subjects(predicate=odmlns.hasUnit, object=Literal(p_unit)))), 1)
        self.assertEqual(len(list(w.g.subjects(predicate=odmlns.hasDefinition, object=Literal(p_def)))), 1)
        self.assertEqual(len(list(w.g.subjects(predicate=odmlns.hasUncertainty, object=Literal(p_uncertainty)))), 1)
        self.assertEqual(len(list(w.g.subjects(predicate=odmlns.hasDtype, object=Literal(p_dtype)))), 1)
        self.assertEqual(len(list(w.g.subjects(predicate=odmlns.hasValueOrigin, object=Literal(p_value_origin)))), 1)
        self.assertEqual(len(list(w.g.subjects(predicate=odmlns.hasReference, object=Literal(p_ref)))), 1)



import os
import tempfile
import unittest
from contextlib import contextmanager

from odml.tools.format_converter import FormatConverter

FC = FormatConverter

# TODO The used NamedTemporaryFile does not play nice with Windows;
# deactivating all affected tests for Windows until this has been fixed.


class TestFormatConverter(unittest.TestCase):
    def setUp(self):
        self.doc = """<odML version="1.1">
                        <section>
                            <type>some</type>
                            <name>S</name>
                        </section>
                      </odML>
                   """

    @contextmanager
    def assertNotRaises(self, exc_type):
        try:
            yield None
        except exc_type:
            raise self.failureException('{} raised'.format(exc_type.__name__))

    def test_convert(self):
        if os.name == 'nt':
            raise unittest.SkipTest("Skipping test on Windows")

        self.test_convert_dir_no_output_dir(False, FC.convert)
        self.test_convert_dir_no_output_dir(True, FC.convert)
        self.test_convert_dir_with_output_dir_specified(FC.convert)

    def test_convert_dir(self):
        if os.name == 'nt':
            raise unittest.SkipTest("Skipping test on Windows")

        with self.assertRaises(ValueError):
            FC.convert_dir(None, None, False, "not valid format")

        # Testing recursive part of the converter for not nested input dir
        self.test_convert_dir_no_output_dir(True)

    def test_convert_dir_no_output_dir(self, recursive=False, func=None):
        if os.name == 'nt':
            raise unittest.SkipTest("Skipping test on Windows")

        work_dir = tempfile.mkdtemp()
        in_dir = tempfile.mkdtemp(dir=work_dir)
        in_file = self.create_open_file(in_dir)
        in_file2 = self.create_open_file(in_dir)

        if not func:
            FC.convert_dir(in_dir, None, recursive, "odml")
        else:
            if recursive:
                func([in_dir, "odml", "-r"])
            else:
                func([in_dir, "odml"])

        files = []
        for dir_path, dir_names, file_names in os.walk(work_dir):
            for file_name in file_names:
                files.append(os.path.join(dir_path, file_name))

        # check if the input file in the correct repo
        self.assertIn(os.path.join(in_dir, in_file.name), files)
        self.assertIn(os.path.join(in_dir, in_file2.name), files)

        # check if the output file in correct repo and has intended name
        in_files = [in_file.name, in_file2.name]
        for file in in_files:
            odml_file_name = os.path.basename(file)
            odml_file_name = odml_file_name.replace(".xml", ".odml")
            root, odml_dir_name = os.path.split(in_dir)
            odml_dir_name = odml_dir_name + "_odml"
            self.assertIn(os.path.join(root, odml_dir_name, odml_file_name), files)

        in_file.close()
        in_file2.close()

    def test_convert_dir_with_output_dir_specified(self, func=None):
        if os.name == 'nt':
            raise unittest.SkipTest("Skipping test on Windows")

        # Testing FC.convert_dir(in_dir, out_dir, False, "odml")
        in_dir = tempfile.mkdtemp()
        out_dir = tempfile.mkdtemp()
        in_file = self.create_open_file(in_dir)
        in_file2 = self.create_open_file(in_dir)

        if not func:
            FC.convert_dir(in_dir, out_dir, False, "odml")
        else:
            func([in_dir, "odml", "-out", out_dir])

        in_files = []
        out_files = []
        for dir_path, dir_names, file_names in os.walk(in_dir):
            for file_name in file_names:
                in_files.append(os.path.join(dir_path, file_name))

        for dir_path, dir_names, file_names in os.walk(out_dir):
            for file_name in file_names:
                out_files.append(os.path.join(dir_path, file_name))

        # check if the input file in the correct repo
        self.assertIn(os.path.join(in_dir, in_file.name), in_files)
        self.assertIn(os.path.join(in_dir, in_file2.name), in_files)

        # check if the output file in correct repo and has intended name
        check_in_files = [in_file.name, in_file2.name]
        for file in check_in_files:
            _, out_file_name = os.path.split(file)
            pre, ext = os.path.splitext(out_file_name)
            out_file_name = out_file_name.replace(out_file_name, pre + ".odml")
            self.assertIn(os.path.join(out_dir, out_file_name), out_files)

        in_file.close()
        in_file2.close()

    def create_open_file(self, in_dir):
        in_file = tempfile.NamedTemporaryFile(mode='a+', suffix=".xml", dir=in_dir)
        in_file.write(self.doc)
        in_file.seek(0)
        return in_file

    def test_check_io_directory(self):
        out_dir = tempfile.mkdtemp()
        in_dir = tempfile.mkdtemp()
        with self.assertRaises(ValueError):
            FC._check_input_output_directory(None, None)
        with self.assertRaises(ValueError):
            FC._check_input_output_directory("/not_valid_path", None)
        with self.assertRaises(ValueError):
            FC._check_input_output_directory(in_dir, "/not_valid_path")
        self.assertNotRaises(FC._check_input_output_directory(in_dir, None))
        self.assertNotRaises(FC._check_input_output_directory(in_dir, out_dir))




import unittest
import sys
import os
import odml

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


class TestTypes(unittest.TestCase):
    # TODO :- Write tests for JSONParser once it's completed.

    def setUp(self):
        self.file = 'doc/example_odMLs/THGTTG.odml'
        # Do not allow anything to be printed on STDOUT
        self.captured_stdout = StringIO()
        sys.stdout = self.captured_stdout

    def test_load_save(self):
        doc = odml.load(self.file)
        self.assertTrue(isinstance(doc, odml.doc.BaseDocument))
        odml.save(doc, self.file + '_copy')
        os.remove(self.file + '_copy')

    def test_display(self):
        doc = odml.load(self.file)
        odml.display(doc)

    def test_invalid_parser(self):
        with self.assertRaises(NotImplementedError):
            odml.load(self.file, 'html')

        doc = odml.load(self.file)
        with self.assertRaises(NotImplementedError):
            odml.save(doc, self.file + '_copy_html', 'html')

        with self.assertRaises(NotImplementedError):
            odml.display(doc, 'html')




"""
This file tests proper creation, saving and loading
of all supported odML datatypes with all supported
odML parsers.
"""

import datetime as dt
import os
import shutil
import tempfile
import unittest

import odml


class TestTypesIntegration(unittest.TestCase):

    def setUp(self):
        # Set up test environment
        self.tmp_dir = tempfile.mkdtemp(suffix=".odml")

        self.json_file = os.path.join(self.tmp_dir, "test.json")
        self.xml_file = os.path.join(self.tmp_dir, "test.xml")
        self.yaml_file = os.path.join(self.tmp_dir, "test.yaml")

        # Set up odML document stub
        doc = odml.Document()
        _ = odml.Section(name="dtypes", type="test", parent=doc)
        self.doc = doc

    def tearDown(self):
        if os.path.exists(self.tmp_dir):
            shutil.rmtree(self.tmp_dir)

    def test_time(self):
        time_string = '12:34:56'
        time = dt.time(12, 34, 56)
        val_in = time_string
        vals_in = [None, "", time_string, time]
        parent_sec = self.doc.sections[0]
        _ = odml.Property(name="time test single", dtype="time",
                          value=val_in, parent=parent_sec)
        _ = odml.Property(name="time test", dtype="time",
                          value=vals_in, parent=parent_sec)

        # Test correct json save and load.
        odml.save(self.doc, self.json_file, "JSON")
        jdoc = odml.load(self.json_file, "JSON")

        self.assertEqual(jdoc.sections[0].properties[0].dtype, odml.dtypes.DType.time)
        self.assertIsInstance(jdoc.sections[0].properties[0].value[0], dt.time)
        self.assertEqual(jdoc.sections[0].properties[1].dtype, odml.dtypes.DType.time)
        for val in jdoc.sections[0].properties[1].value:
            self.assertIsInstance(val, dt.time)
        self.assertEqual(jdoc.sections[0].properties[1].value[2], time)
        self.assertEqual(jdoc.sections[0].properties[1].value[3], time)
        self.assertEqual(self.doc, jdoc)

        # Test correct xml save and load.
        odml.save(self.doc, self.xml_file)
        xdoc = odml.load(self.xml_file)

        self.assertEqual(xdoc.sections[0].properties[0].dtype, odml.dtypes.DType.time)
        self.assertIsInstance(xdoc.sections[0].properties[0].value[0], dt.time)
        self.assertEqual(xdoc.sections[0].properties[1].dtype, odml.dtypes.DType.time)
        for val in xdoc.sections[0].properties[1].value:
            self.assertIsInstance(val, dt.time)
        self.assertEqual(xdoc.sections[0].properties[1].value[2], time)
        self.assertEqual(xdoc.sections[0].properties[1].value[2], time)
        self.assertEqual(self.doc, xdoc)

        # Test correct yaml save and load.
        odml.save(self.doc, self.yaml_file, "YAML")
        ydoc = odml.load(self.yaml_file, "YAML")

        self.assertEqual(ydoc.sections[0].properties[0].dtype, odml.dtypes.DType.time)
        self.assertIsInstance(ydoc.sections[0].properties[0].value[0], dt.time)
        self.assertEqual(ydoc.sections[0].properties[1].dtype, odml.dtypes.DType.time)
        for val in ydoc.sections[0].properties[1].value:
            self.assertIsInstance(val, dt.time)
        self.assertEqual(ydoc.sections[0].properties[1].value[2], time)
        self.assertEqual(ydoc.sections[0].properties[1].value[2], time)
        self.assertEqual(self.doc, ydoc)

    def test_date(self):
        date_string = '2018-08-31'
        date = dt.date(2018, 8, 31)
        val_in = date_string
        vals_in = [None, "", date_string, date]
        parent_sec = self.doc.sections[0]
        _ = odml.Property(name="date test single", dtype="date",
                          value=val_in, parent=parent_sec)
        _ = odml.Property(name="date test", dtype="date",
                          value=vals_in, parent=parent_sec)

        # Test correct json save and load.
        odml.save(self.doc, self.json_file, "JSON")
        jdoc = odml.load(self.json_file, "JSON")

        self.assertEqual(jdoc.sections[0].properties[0].dtype, odml.dtypes.DType.date)
        self.assertIsInstance(jdoc.sections[0].properties[0].value[0], dt.date)
        self.assertEqual(jdoc.sections[0].properties[1].dtype, odml.dtypes.DType.date)
        for val in jdoc.sections[0].properties[1].value:
            self.assertIsInstance(val, dt.date)
        self.assertEqual(jdoc.sections[0].properties[1].value[2], date)
        self.assertEqual(jdoc.sections[0].properties[1].value[3], date)
        self.assertEqual(self.doc, jdoc)

        # Test correct xml save and load.
        odml.save(self.doc, self.xml_file)
        xdoc = odml.load(self.xml_file)

        self.assertEqual(xdoc.sections[0].properties[0].dtype, odml.dtypes.DType.date)
        self.assertIsInstance(xdoc.sections[0].properties[0].value[0], dt.date)
        self.assertEqual(xdoc.sections[0].properties[1].dtype, odml.dtypes.DType.date)
        for val in xdoc.sections[0].properties[1].value:
            self.assertIsInstance(val, dt.date)
        self.assertEqual(xdoc.sections[0].properties[1].value[2], date)
        self.assertEqual(xdoc.sections[0].properties[1].value[2], date)
        self.assertEqual(self.doc, xdoc)

        # Test correct yaml save and load.
        odml.save(self.doc, self.yaml_file, "YAML")
        ydoc = odml.load(self.yaml_file, "YAML")

        self.assertEqual(ydoc.sections[0].properties[0].dtype, odml.dtypes.DType.date)
        self.assertIsInstance(ydoc.sections[0].properties[0].value[0], dt.date)
        self.assertEqual(ydoc.sections[0].properties[1].dtype, odml.dtypes.DType.date)
        for val in ydoc.sections[0].properties[1].value:
            self.assertIsInstance(val, dt.date)
        self.assertEqual(ydoc.sections[0].properties[1].value[2], date)
        self.assertEqual(ydoc.sections[0].properties[1].value[2], date)
        self.assertEqual(self.doc, ydoc)

    def test_datetime(self):
        datetime_string = '2018-08-31 12:59:59'
        date_time = dt.datetime(2018, 8, 31, 12, 59, 59)
        val_in = datetime_string
        vals_in = [None, "", datetime_string, date_time]
        parent_sec = self.doc.sections[0]
        _ = odml.Property(name="datetime test single", dtype="datetime",
                          value=val_in, parent=parent_sec)
        _ = odml.Property(name="datetime test", dtype="datetime",
                          value=vals_in, parent=parent_sec)

        # Test correct json save and load.
        odml.save(self.doc, self.json_file, "JSON")
        jdoc = odml.load(self.json_file, "JSON")

        self.assertEqual(jdoc.sections[0].properties[0].dtype, odml.dtypes.DType.datetime)
        self.assertIsInstance(jdoc.sections[0].properties[0].value[0], dt.datetime)
        self.assertEqual(jdoc.sections[0].properties[1].dtype, odml.dtypes.DType.datetime)
        for val in jdoc.sections[0].properties[1].value:
            self.assertIsInstance(val, dt.datetime)
        self.assertEqual(jdoc.sections[0].properties[1].value[2], date_time)
        self.assertEqual(jdoc.sections[0].properties[1].value[3], date_time)
        self.assertEqual(self.doc, jdoc)

        # Test correct xml save and load.
        odml.save(self.doc, self.xml_file)
        xdoc = odml.load(self.xml_file)

        self.assertEqual(xdoc.sections[0].properties[0].dtype, odml.dtypes.DType.datetime)
        self.assertIsInstance(xdoc.sections[0].properties[0].value[0], dt.datetime)
        self.assertEqual(xdoc.sections[0].properties[1].dtype, odml.dtypes.DType.datetime)
        for val in xdoc.sections[0].properties[1].value:
            self.assertIsInstance(val, dt.datetime)
        self.assertEqual(xdoc.sections[0].properties[1].value[2], date_time)
        self.assertEqual(xdoc.sections[0].properties[1].value[2], date_time)
        self.assertEqual(self.doc, xdoc)

        # Test correct yaml save and load.
        odml.save(self.doc, self.yaml_file, "YAML")
        ydoc = odml.load(self.yaml_file, "YAML")

        self.assertEqual(ydoc.sections[0].properties[0].dtype, odml.dtypes.DType.datetime)
        self.assertIsInstance(ydoc.sections[0].properties[0].value[0], dt.datetime)
        self.assertEqual(ydoc.sections[0].properties[1].dtype, odml.dtypes.DType.datetime)
        for val in ydoc.sections[0].properties[1].value:
            self.assertIsInstance(val, dt.datetime)
        self.assertEqual(ydoc.sections[0].properties[1].value[2], date_time)
        self.assertEqual(ydoc.sections[0].properties[1].value[2], date_time)
        self.assertEqual(self.doc, ydoc)

    def test_int(self):
        val_in = [1, 2, 3, "4"]
        val_odml = [1, 2, 3, 4]
        parent_sec = self.doc.sections[0]
        _ = odml.Property(name="int test", dtype="int", value=val_in, parent=parent_sec)

        # Test correct json save and load.
        odml.save(self.doc, self.json_file, "JSON")
        jdoc = odml.load(self.json_file, "JSON")

        self.assertEqual(jdoc.sections[0].properties[0].dtype, odml.dtypes.DType.int)
        self.assertEqual(jdoc.sections[0].properties[0].value, val_odml)
        self.assertEqual(self.doc, jdoc)

        # Test correct xml save and load.
        odml.save(self.doc, self.xml_file)
        xdoc = odml.load(self.xml_file)

        self.assertEqual(xdoc.sections[0].properties[0].dtype, odml.dtypes.DType.int)
        self.assertEqual(xdoc.sections[0].properties[0].value, val_odml)
        self.assertEqual(self.doc, xdoc)

        # Test correct yaml save and load.
        odml.save(self.doc, self.yaml_file, "YAML")
        ydoc = odml.load(self.yaml_file, "YAML")

        self.assertEqual(ydoc.sections[0].properties[0].dtype, odml.dtypes.DType.int)
        self.assertEqual(ydoc.sections[0].properties[0].value, val_odml)
        self.assertEqual(self.doc, ydoc)

    def test_float(self):
        val_in = [1, 2.2, 3.3, "4"]
        val_odml = [1.0, 2.2, 3.3, 4.0]
        parent_sec = self.doc.sections[0]
        _ = odml.Property(name="float test", dtype="float",
                          value=val_in, parent=parent_sec)

        # Test correct json save and load.
        odml.save(self.doc, self.json_file, "JSON")
        jdoc = odml.load(self.json_file, "JSON")

        self.assertEqual(jdoc.sections[0].properties[0].dtype, odml.dtypes.DType.float)
        self.assertEqual(jdoc.sections[0].properties[0].value, val_odml)
        self.assertEqual(self.doc, jdoc)

        # Test correct xml save and load.
        odml.save(self.doc, self.xml_file)
        xdoc = odml.load(self.xml_file)

        self.assertEqual(xdoc.sections[0].properties[0].dtype, odml.dtypes.DType.float)
        self.assertEqual(xdoc.sections[0].properties[0].value, val_odml)
        self.assertEqual(self.doc, xdoc)

        # Test correct yaml save and load.
        odml.save(self.doc, self.yaml_file, "YAML")
        ydoc = odml.load(self.yaml_file, "YAML")

        self.assertEqual(ydoc.sections[0].properties[0].dtype, odml.dtypes.DType.float)
        self.assertEqual(ydoc.sections[0].properties[0].value, val_odml)
        self.assertEqual(self.doc, ydoc)

    def test_str(self):
        val_in = "single value"
        vals_in = [None, "", [], {}, 1, True, "text"]
        vals_odml = ["", "", "", "", "1", "True", "text"]
        parent_sec = self.doc.sections[0]
        _ = odml.Property(name="string test single", dtype="string",
                          value=val_in, parent=parent_sec)
        _ = odml.Property(name="string test", dtype="string",
                          value=vals_in, parent=parent_sec)

        # Test correct json save and load.
        odml.save(self.doc, self.json_file, "JSON")
        jdoc = odml.load(self.json_file, "JSON")

        self.assertEqual(jdoc.sections[0].properties[0].dtype, odml.dtypes.DType.string)
        self.assertEqual(jdoc.sections[0].properties[0].value, [val_in])
        self.assertEqual(jdoc.sections[0].properties[1].dtype, odml.dtypes.DType.string)
        self.assertEqual(jdoc.sections[0].properties[1].value, vals_odml)
        self.assertEqual(self.doc, jdoc)

        # Test correct xml save and load.
        odml.save(self.doc, self.xml_file)
        xdoc = odml.load(self.xml_file)

        self.assertEqual(xdoc.sections[0].properties[0].dtype, odml.dtypes.DType.string)
        self.assertEqual(xdoc.sections[0].properties[0].value, [val_in])
        self.assertEqual(xdoc.sections[0].properties[1].dtype, odml.dtypes.DType.string)
        self.assertEqual(xdoc.sections[0].properties[1].value, vals_odml)
        self.assertEqual(self.doc, xdoc)

        # Test correct yaml save and load.
        odml.save(self.doc, self.yaml_file, "YAML")
        ydoc = odml.load(self.yaml_file, "YAML")

        self.assertEqual(ydoc.sections[0].properties[0].dtype, odml.dtypes.DType.string)
        self.assertEqual(ydoc.sections[0].properties[0].value, [val_in])
        self.assertEqual(ydoc.sections[0].properties[1].dtype, odml.dtypes.DType.string)
        self.assertEqual(ydoc.sections[0].properties[1].value, vals_odml)
        self.assertEqual(self.doc, ydoc)

    def test_bool(self):
        val_in = True
        vals_in = [None, "", [], {}, False, True, "TRUE"]
        vals_odml = [False, False, False, False, False, True, True]
        parent_sec = self.doc.sections[0]
        _ = odml.Property(name="bool test single", dtype="boolean",
                          value=val_in, parent=parent_sec)
        _ = odml.Property(name="bool test", dtype="boolean",
                          value=vals_in, parent=parent_sec)

        # Test correct json save and load.
        odml.save(self.doc, self.json_file, "JSON")
        jdoc = odml.load(self.json_file, "JSON")

        self.assertEqual(jdoc.sections[0].properties[0].dtype, odml.dtypes.DType.boolean)
        self.assertEqual(jdoc.sections[0].properties[0].value, [val_in])
        self.assertEqual(jdoc.sections[0].properties[1].dtype, odml.dtypes.DType.boolean)
        self.assertEqual(jdoc.sections[0].properties[1].value, vals_odml)
        self.assertEqual(self.doc, jdoc)

        # Test correct xml save and load.
        odml.save(self.doc, self.xml_file)
        xdoc = odml.load(self.xml_file)

        self.assertEqual(xdoc.sections[0].properties[0].dtype, odml.dtypes.DType.boolean)
        self.assertEqual(xdoc.sections[0].properties[0].value, [val_in])
        self.assertEqual(xdoc.sections[0].properties[1].dtype, odml.dtypes.DType.boolean)
        self.assertEqual(xdoc.sections[0].properties[1].value, vals_odml)
        self.assertEqual(self.doc, xdoc)

        # Test correct yaml save and load.
        odml.save(self.doc, self.yaml_file, "YAML")
        ydoc = odml.load(self.yaml_file, "YAML")

        self.assertEqual(ydoc.sections[0].properties[0].dtype, odml.dtypes.DType.boolean)
        self.assertEqual(ydoc.sections[0].properties[0].value, [val_in])
        self.assertEqual(ydoc.sections[0].properties[1].dtype, odml.dtypes.DType.boolean)
        self.assertEqual(ydoc.sections[0].properties[1].value, vals_odml)
        self.assertEqual(self.doc, ydoc)

    def test_tuple(self):
        val_type = "3-tuple"
        val_in = "(1; 1; 1)"
        val_odml = ["1", "1", "1"]

        parent_sec = self.doc.sections[0]
        _ = odml.Property(name="tuple test single", dtype=val_type,
                          value=val_in, parent=parent_sec)

        # Test correct json save and load.
        odml.save(self.doc, self.json_file, "JSON")
        jdoc = odml.load(self.json_file, "JSON")

        self.assertEqual(jdoc.sections[0].properties[0].dtype, val_type)
        self.assertEqual(jdoc.sections[0].properties[0].value, [val_odml])
        self.assertEqual(self.doc, jdoc)

        # Test correct xml save and load.
        odml.save(self.doc, self.xml_file)
        xdoc = odml.load(self.xml_file)

        self.assertEqual(xdoc.sections[0].properties[0].dtype, val_type)
        self.assertEqual(xdoc.sections[0].properties[0].value, [val_odml])
        self.assertEqual(self.doc, xdoc)

        # Test correct yaml save and load.
        odml.save(self.doc, self.yaml_file, "YAML")
        ydoc = odml.load(self.yaml_file, "YAML")

        self.assertEqual(ydoc.sections[0].properties[0].dtype, val_type)
        self.assertEqual(ydoc.sections[0].properties[0].value, [val_odml])
        self.assertEqual(self.doc, ydoc)



#!/usr/bin/env python
"""
The XML parsing module.
Parses odML files. Can be invoked standalone:
    python -m odml.tools.xmlparser file.odml
"""
import csv
import sys
from lxml import etree as ET
from lxml.builder import E
# this is needed for py2exe to include lxml completely
from lxml import _elementpath as _dummy
from os.path import basename

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from .. import format
from ..info import FORMAT_VERSION
from .parser_utils import InvalidVersionException, ParserException

try:
    unicode = unicode
except NameError:
    unicode = str


def to_csv(val):
    # Make sure all individual values do not contain
    # leading or trailing whitespaces.
    unicode_values = list(map(unicode.strip, map(unicode, val)))
    stream = StringIO()
    writer = csv.writer(stream, dialect="excel")
    writer.writerow(unicode_values)
    # Strip any csv.writer added carriage return line feeds
    # and double quotes before saving.
    csv_string = stream.getvalue().strip().strip('"')
    if len(unicode_values) > 1:
        csv_string = "[" + csv_string + "]"
    return csv_string


def from_csv(value_string):
    if not value_string:
        return []
    if value_string[0] == "[" and value_string[-1] == "]":
        value_string = value_string[1:-1]
    else:
        # This is a single string entry, any comma contained
        # is part of the value and must not be used to
        # split up the string.
        return [value_string]

    if not value_string:
        return []
    stream = StringIO(value_string)
    stream.seek(0)
    reader = csv.reader(stream, dialect="excel")
    return list(reader)[0]


class XMLWriter:
    """
    Creates XML nodes storing the information of an odML Document
    """
    header = """<?xml version="1.0" encoding="UTF-8"?>
<?xml-stylesheet  type="text/xsl" href="odmlTerms.xsl"?>
<?xml-stylesheet  type="text/xsl" href="odml.xsl"?>
"""

    def __init__(self, odml_document):
        self.doc = odml_document

    @staticmethod
    def save_element(e):
        """
        returns an xml node for the odML object e
        """
        fmt = e.format()
        cur = E(fmt.name)

        # generate attributes
        if isinstance(fmt, format.Document.__class__):
            cur.attrib['version'] = FORMAT_VERSION

        # generate elements
        for k in fmt.arguments_keys:
            if not hasattr(e, fmt.map(k)):
                continue

            val = getattr(e, fmt.map(k))
            if val is None:
                continue
            if isinstance(fmt, format.Property.__class__) and k == "value":
                # Custom odML tuples require special handling for save loading from file.
                if e.dtype and e.dtype.endswith("-tuple") and len(val) > 0:
                    ele = E(k, "(%s)" % ";".join(val[0]))
                else:
                    ele = E(k, to_csv(val))
                cur.append(ele)
            else:
                if isinstance(val, list):
                    for v in val:
                        if v is None:
                            continue
                        ele = XMLWriter.save_element(v)
                        cur.append(ele)
                else:
                    if sys.version_info < (3,):
                        ele = E(k, unicode(val))
                    else:
                        ele = E(k, str(val))
                    cur.append(ele)
        return cur

    def __str__(self):
        return ET.tounicode(self.save_element(self.doc), pretty_print=True)

    def __unicode__(self):
        return ET.tounicode(self.save_element(self.doc), pretty_print=True)

    def write_file(self, filename):
        # calculate the data before opening the file in case we get any
        # exception
        if sys.version_info < (3,):
            data = unicode(self).encode('utf-8')
        else:
            data = str(self)

        with open(filename, "w") as file:
            file.write(self.header)
            file.write(data)


def load(filename):
    """
    shortcut function for XMLReader().from_file(filename)
    """
    return XMLReader().from_file(filename)


class XMLReader(object):
    """
    A reader to parse xml-files or strings into odml data structures
    Usage:
        >>> doc = XMLReader().from_file("file.odml")
    """

    def __init__(self, ignore_errors=False, filename=None):
        self.parser = ET.XMLParser(remove_comments=True)
        self.tags = dict([(obj.name, obj) for obj in format.__all__])
        self.ignore_errors = ignore_errors
        self.filename = filename
        self.warnings = []

    @staticmethod
    def _handle_version(root):
        """
        Check if the odML version of a handed in parsed lxml.etree is supported
        by the current library and raise an Exception otherwise.
        :param root: Root node of a parsed lxml.etree. The root tag has to
                     contain a supported odML version number, otherwise it is not
                     accepted as a valid odML file.
        """
        if root.tag != 'odML':
            raise ParserException("Expecting <odML> tag but got <%s>.\n" % root.tag)
        elif 'version' not in root.attrib:
            raise ParserException("Could not find format version attribute "
                                  "in <odML> tag.\n")
        elif root.attrib['version'] != FORMAT_VERSION:
            msg = ("Cannot parse odML document with format version '%s'. \n"
                   "\tUse the 'tools.VersionConverter' to import previous odML formats."
                   % root.attrib['version'])
            raise InvalidVersionException(msg)

    def from_file(self, xml_file):
        """
        parse the datastream from a file like object *xml_file*
        and return an odml data structure
        """
        try:
            root = ET.parse(xml_file, self.parser).getroot()
            if hasattr(xml_file, "close"):
                xml_file.close()
        except ET.XMLSyntaxError as e:
            raise ParserException(e.msg)

        self._handle_version(root)
        doc = self.parse_element(root)

        # Provide original file name via the in memory document
        if isinstance(xml_file, unicode):
            doc._origin_file_name = basename(xml_file)
        return doc

    def from_string(self, string):
        try:
            root = ET.XML(string, self.parser)
        except ET.XMLSyntaxError as e:
            raise ParserException(e.msg)

        self._handle_version(root)
        return self.parse_element(root)

    def check_mandatory_arguments(self, data, ArgClass, tag_name, node):
        for k, v in ArgClass.arguments:
            if v != 0 and not ArgClass.map(k) in data:
                self.error("missing element <%s> within <%s> tag" %
                           (k, tag_name) + repr(data), node)

    def is_valid_argument(self, tag_name, ArgClass, parent_node, child=None):
        if tag_name not in ArgClass.arguments_keys:
            self.error("Invalid element <%s> inside <%s> tag" %
                       (tag_name, parent_node.tag),
                       parent_node if child is None else child)

    def error(self, msg, elem):
        if elem is not None:
            msg += " (line %d)" % elem.sourceline
        if self.ignore_errors:
            return self.warn(msg, elem)
        raise ParserException(msg)

    def warn(self, msg, elem):
        if elem is not None:
            msg = "warning[%s:%d:<%s>]: %s\n" % (
                self.filename, elem.sourceline, elem.tag, msg)
        else:
            msg = "warning: %s\n" % msg
        self.warnings.append(msg)
        sys.stderr.write(msg)

    def parse_element(self, node):
        if node.tag not in self.tags:
            self.error("Invalid element <%s>" % node.tag, node)
            return None  # won't be able to parse this one
        return getattr(self, "parse_" + node.tag)(node, self.tags[node.tag])

    def parse_tag(self, root, fmt, insert_children=True):
        """
        Parse an odml node based on the format description *fmt*
        and instantiate the corresponding object.
        :param root: lxml.etree node containing an odML object or object tree.
        :param fmt: odML class corresponding to the content of the root node.
        :param insert_children: Bool value. When True, child elements of the root node
                                will be parsed to their odML equivalents and appended to
                                the odML document. When False, child elements of the
                                root node will be ignored.
        """
        arguments = {}
        extra_args = {}
        children = []

        for k, v in root.attrib.iteritems():
            k = k.lower()
            # 'version' is currently the only supported XML attribute.
            if k == 'version' and root.tag == 'odML':
                continue

            # We currently do not support XML attributes.
            self.error("Attribute not supported, ignoring '%s=%s'" % (k, v), root)

        for node in root:
            node.tag = node.tag.lower()
            self.is_valid_argument(node.tag, fmt, root, node)
            if node.tag in fmt.arguments_keys:
                # this is a heuristic, but works for now
                if node.tag in self.tags and node.tag in fmt.map_keys:
                    sub_obj = self.parse_element(node)
                    if sub_obj is not None:
                        extra_args[fmt.map(node.tag)] = sub_obj
                        children.append(sub_obj)
                else:
                    tag = fmt.map(node.tag)
                    if tag in arguments:
                        self.warn("Element <%s> is given multiple times in "
                                  "<%s> tag" % (node.tag, root.tag), node)

                    # Special handling of values;
                    curr_text = node.text.strip() if node.text else None
                    if tag == "value" and curr_text:
                        content = from_csv(node.text)
                        arguments[tag] = content
                    else:
                        arguments[tag] = curr_text
            else:
                self.error("Invalid element <%s> in odML document section <%s>"
                           % (node.tag, root.tag), node)

        if sys.version_info > (3,):
            check_args = dict(list(arguments.items()) + list(extra_args.items()))
        else:
            check_args = dict(arguments.items() + extra_args.items())

        self.check_mandatory_arguments(check_args, fmt, root.tag, root)

        # Instantiate the current odML object with the parsed attributes.
        obj = fmt.create(**arguments)

        if insert_children:
            for child in children:
                obj.append(child)

        return obj

    def parse_odML(self, root, fmt):
        doc = self.parse_tag(root, fmt)
        return doc

    def parse_section(self, root, fmt):
        return self.parse_tag(root, fmt)

    def parse_property(self, root, fmt):
        return self.parse_tag(root, fmt, insert_children=False)


if __name__ == '__main__':
    from optparse import OptionParser
    import odml.tools.dumper as dumper

    parser = OptionParser()
    (options, args) = parser.parse_args()

    if len(args) < 1:
        parser.print_help()
    else:
        dumper.dumpDoc(load(args[0]))



"""
Handles (deferred) loading of terminology data and access to it
for odML documents
"""

import datetime
import os
import sys
import tempfile
import threading
try:
    import urllib.request as urllib2
except ImportError:
    import urllib2

from hashlib import md5

from .tools.parser_utils import ParserException
from .tools.xmlparser import XMLReader


REPOSITORY_BASE = 'http://portal.g-node.org/odml/terminologies'
REPOSITORY = '/'.join([REPOSITORY_BASE, 'v1.1', 'terminologies.xml'])

CACHE_AGE = datetime.timedelta(days=1)


def cache_load(url):
    """
    Load the url and store it in a temporary cache directory
    subsequent requests for this url will use the cached version
    """
    filename = '.'.join([md5(url.encode()).hexdigest(), os.path.basename(url)])
    cache_dir = os.path.join(tempfile.gettempdir(), "odml.cache")
    if not os.path.exists(cache_dir):
        try:
            os.makedirs(cache_dir)
        except OSError:  # might happen due to concurrency
            if not os.path.exists(cache_dir):
                raise
    cache_file = os.path.join(cache_dir, filename)
    if not os.path.exists(cache_file) \
       or datetime.datetime.fromtimestamp(os.path.getmtime(cache_file)) < \
       datetime.datetime.now() - CACHE_AGE:
        try:
            data = urllib2.urlopen(url).read()
            if sys.version_info.major > 2:
                data = data.decode("utf-8")
        except Exception as e:
            print("failed loading '%s': %s" % (url, e))
            return
        fp = open(cache_file, "w")
        fp.write(str(data))
        fp.close()
    return open(cache_file)


class Terminologies(dict):
    loading = {}

    def load(self, url):
        """
        Load and cache a terminology-url

        Returns the odml-document for the url
        """
        if url in self:
            return self[url]

        if url in self.loading:
            self.loading[url].join()
            self.loading.pop(url, None)
            return self.load(url)

        return self._load(url)

    def _load(self, url):
        # TODO also cache the data locally on disk
        # if url.startswith("http"): return None
        fp = cache_load(url)
        if fp is None:
            print("did not successfully load '%s'" % url)
            return
        try:
            term = XMLReader(filename=url, ignore_errors=True).from_file(fp)
            term.finalize()
        except ParserException as e:
            print("Failed to load %s due to parser errors" % url)
            print(' "%s"' % e)
            term = None
        self[url] = term
        return term

    def deferred_load(self, url):
        """
        Start a thread to load the terminology in background
        """
        if url in self or url in self.loading:
            return
        self.loading[url] = threading.Thread(target=self._load, args=(url,))
        self.loading[url].start()


terminologies = Terminologies()
load = terminologies.load
deferred_load = terminologies.deferred_load


if __name__ == "__main__":
    f = cache_load(REPOSITORY)



# -*- coding: utf-8
"""
Collects common base functionality
"""
import collections
import posixpath

from . import terminology
from .tools.doc_inherit import allow_inherit_docstring


class BaseObject(object):
    _format = None

    def __hash__(self):
        """
        Allow all odML objects to be hash-able.
        """
        return id(self)

    def __eq__(self, obj):
        """
        Do a deep comparison of this object and its odml properties.
        The 'id' attribute of an object is excluded, since it is
        unique within a document.
        """
        # cannot compare totally different stuff
        if not isinstance(self, obj.__class__):
            return False

        for key in self._format:
            if key == "id" or key == "oid":
                continue
            elif getattr(self, key) != getattr(obj, key):
                return False

        return True

    def __ne__(self, obj):
        """
        Use the __eq__ function to determine if both objects are equal
        """
        return not self == obj

    def format(self):
        return self._format

    @property
    def document(self):
        """ Returns the Document object in which this object is contained """
        if self.parent is None:
            return None
        return self.parent.document

    def get_terminology_equivalent(self):
        """
        Returns the equivalent object in an terminology (should there be one
        defined) or None
        """
        return None

    def clean(self):
        """
        Stub that doesn't do anything for this class
        """
        pass

    def clone(self, children=True):
        """
        Clone this object recursively (if children is True) allowing to copy it
        independently to another document. If children is False, this acts as
        a template cloner, creating a copy of the object without any children
        """
        # TODO don't we need some recursion / deepcopy here?
        import copy
        obj = copy.copy(self)
        return obj


class SmartList(list):

    def __init__(self, content_type):
        """
        Only values of the instance *content_type* can be added to the SmartList.
        """
        self._content_type = content_type
        super(SmartList, self).__init__()

    def __getitem__(self, key):
        """
        Provides element index also by searching for an element with a given name.
        """
        # Try normal list index first (for integers)
        if isinstance(key, int):
            return super(SmartList, self).__getitem__(key)

        # Otherwise search the list
        for obj in self:
            if (hasattr(obj, "name") and obj.name == key) or key == obj:
                return obj

        # and fail eventually
        raise KeyError(key)

    def __setitem__(self, key, value):
        """
        Replaces item at list[*key*] with *value*.
        :param key: index position
        :param value: object that replaces item at *key* position.
                      value has to be of the same content type as the list.
                      In this context usually a Section or a Property.
        """
        if not isinstance(value, self._content_type):
            raise ValueError("List only supports elements of type '%s'" %
                             self._content_type)

        # If required remove new object from its old parents child-list
        if hasattr(value, "_parent") and (value._parent and value in value._parent):
            value._parent.remove(value)

        # If required move parent reference from replaced to new object
        # and set parent reference on replaced object None.
        if hasattr(self[key], "_parent"):
            value._parent = self[key]._parent
            self[key]._parent = None

        super(SmartList, self).__setitem__(key, value)

    def __contains__(self, key):
        for obj in self:
            if (hasattr(obj, "name") and obj.name == key) or key == obj:
                return True

    def __eq__(self, obj):
        """
        SmartList attributes of 'sections' and 'properties' are
        handled specially: We want to make sure that the lists'
        objects are properly compared without changing the order
        of the individual lists.
        """
        # This special case was introduced only due to the fact
        # that RDF files will be loaded with randomized list
        # order. With any other file format the list order
        # remains unchanged.
        if sorted(self, key=lambda x: x.name) != sorted(obj, key=lambda x: x.name):
            return False

        return True

    def __ne__(self, obj):
        """
        Use the __eq__ function to determine if both objects are equal
        """
        return not self == obj

    def index(self, obj):
        """
        Find obj in list
        """
        for i, e in enumerate(self):
            if e is obj:
                return i
        raise ValueError("remove: %s not in list" % repr(obj))

    def remove(self, obj):
        """
        Remove an element from this list.
        """
        del self[self.index(obj)]

    def append(self, *obj_tuple):
        for obj in obj_tuple:
            if obj.name in self:
                raise KeyError(
                    "Object with the same name already exists! " + str(obj))

            if not isinstance(obj, self._content_type):
                raise ValueError("List only supports elements of type '%s'" %
                                 self._content_type)

            super(SmartList, self).append(obj)

    def sort(self, key=lambda x: x.name, reverse=False):
        """
        If not otherwise defined, sort by the *name* attribute
        of the lists *_content_type* object.
        """
        super(SmartList, self).sort(key=key, reverse=reverse)


@allow_inherit_docstring
class Sectionable(BaseObject):
    def __init__(self):
        from odml.section import BaseSection
        self._sections = SmartList(BaseSection)
        self._repository = None

    def __getitem__(self, key):
        return self._sections[key]

    def __len__(self):
        return len(self._sections)

    def __iter__(self):
        return self._sections.__iter__()

    @property
    def document(self):
        """
        Returns the parent-most node (if its a document instance) or None
        """
        from odml.doc import BaseDocument
        p = self
        while p.parent:
            p = p.parent
        if isinstance(p, BaseDocument):
            return p

    @property
    def sections(self):
        """ The list of sections contained in this section/document """
        return self._sections

    def insert(self, position, section):
        """
        Insert a Section at the child-list position. A ValueError will be raised,
        if a Section with the same name already exists in the child-list.

        :param position: index at which the object should be inserted.
        :param section: odML Section object.
        """
        from odml.section import BaseSection
        if isinstance(section, BaseSection):
            if section.name in self._sections:
                raise ValueError("Section with name '%s' already exists." % section.name)

            self._sections.insert(position, section)
            section._parent = self
        else:
            raise ValueError("Can only insert objects of type Section.")

    def append(self, section):
        """
        Method appends a single Section to the section child-lists of the current Object.

        :param section: odML Section object.
        """
        from odml.section import BaseSection
        if isinstance(section, BaseSection):
            self._sections.append(section)
            section._parent = self
        elif isinstance(section, collections.Iterable) and not isinstance(section, str):
            raise ValueError("Use extend to add a list of Sections.")
        else:
            raise ValueError("Can only append objects of type Section.")

    def extend(self, sec_list):
        """
        Method adds Sections to the section child-list of the current object.

        :param sec_list: Iterable containing odML Section entries.
        """
        from odml.section import BaseSection
        if not isinstance(sec_list, collections.Iterable):
            raise TypeError("'%s' object is not iterable" % type(sec_list).__name__)

        # Make sure only Sections with unique names will be added.
        for sec in sec_list:
            if not isinstance(sec, BaseSection):
                raise ValueError("Can only extend objects of type Section.")

            elif isinstance(sec, BaseSection) and sec.name in self._sections:
                raise KeyError("Section with name '%s' already exists." % sec.name)

        for sec in sec_list:
            self.append(sec)

    def remove(self, section):
        """ Removes the specified child-section """
        self._sections.remove(section)
        section._parent = None

    def itersections(self, recursive=True, yield_self=False,
                     filter_func=lambda x: True, max_depth=None):
        """
        Iterate each child section

        Example: return all subsections which name contains "foo"
        >>> filter_func = lambda x: getattr(x, 'name').find("foo") > -1
        >>> sec_or_doc.itersections(filter_func=filter_func)

        :param recursive: iterate all child sections recursively (deprecated)
        :type recursive: bool

        :param yield_self: includes itself in the iteration
        :type yield_self: bool

        :param filter_func: accepts a function that will be applied to each
                            iterable. Yields iterable if function returns True
        :type filter_func: function
        """
        stack = []
        # Below: never yield self if self is a Document
        if self == self.document and ((max_depth is None) or (max_depth > 0)):
            for sec in self.sections:
                stack.append((sec, 1))  # (<section>, <level in a tree>)
        elif self != self.document:
            stack.append((self, 0))  # (<section>, <level in a tree>)

        while len(stack) > 0:
            (sec, level) = stack.pop(0)
            if filter_func(sec) and (yield_self if level == 0 else True):
                yield sec

            if max_depth is None or level < max_depth:
                for sec in sec.sections:
                    stack.append((sec, level + 1))

    def iterproperties(self, max_depth=None, filter_func=lambda x: True):
        """
        Iterate each related property (recursively)

        Example: return all children properties which name contains "foo"
        >>> filter_func = lambda x: getattr(x, 'name').find("foo") > -1
        >>> sec_or_doc.iterproperties(filter_func=filter_func)

        :param max_depth: iterate all properties recursively if None, only to
                          a certain level otherwise
        :type max_depth: bool

        :param filter_func: accepts a function that will be applied to each
                            iterable. Yields iterable if function returns True
        :type filter_func: function
        """
        for sec in [s for s in self.itersections(max_depth=max_depth,
                                                 yield_self=True)]:
            if hasattr(sec, "properties"):  # not to fail if odml.Document
                for i in sec.properties:
                    if filter_func(i):
                        yield i

    def itervalues(self, max_depth=None, filter_func=lambda x: True):
        """
        Iterate each related value (recursively)

        # Example: return all children values which string converted version
                   has "foo"
        >>> filter_func = lambda x: str(x).find("foo") > -1
        >>> sec_or_doc.itervalues(filter_func=filter_func)

        :param max_depth: iterate all properties recursively if None, only to
                          a certain level otherwise
        :type max_depth: bool

        :param filter_func: accepts a function that will be applied to each
                            iterable. Yields iterable if function returns True
        :type filter_func: function
        """
        for prop in [p for p in self.iterproperties(max_depth=max_depth)]:
            if filter_func(prop.value):
                yield prop.value

    def contains(self, obj):
        """
        Checks if a subsection of name&type of *obj* is a child of this section
        if so return this child
        """
        for i in self._sections:
            if obj.name == i.name and obj.type == i.type:
                return i

    def _matches(self, obj, key=None, otype=None, include_subtype=False):
        """
        Find out
        * if the *key* matches obj.name (if key is not None)
        * or if *otype* matches obj.type (if type is not None)
        * if type does not match exactly, test for subtype.
        (e.g.stimulus/white_noise)
        comparisons are case-insensitive, however both key and type
        MUST be lower-case.
        """
        name_match = (key is None or (
            key is not None and hasattr(obj, "name") and obj.name == key))

        exact_type_match = (otype is None or (otype is not None and
                                              hasattr(obj, "type") and
                                              obj.type.lower() == otype))

        if not include_subtype:
            return name_match and exact_type_match

        subtype_match = (otype is None or
                         (otype is not None and hasattr(obj, "type") and
                          otype in obj.type.lower().split('/')[:-1]))

        return name_match and (exact_type_match or subtype_match)

    def get_section_by_path(self, path):
        """
        Find a Section through a path like "../name1/name2"

        :param path: path like "../name1/name2"
        :type path: str
        """
        return self._get_section_by_path(path)

    def get_property_by_path(self, path):
        """
        Find a Property through a path like "../name1/name2:property_name"

        :param path: path like "../name1/name2:property_name"
        :type path: str
        """
        laststep = path.split(":")  # assuming section names do not contain :
        found = self._get_section_by_path(laststep[0])
        return self._match_iterable(found.properties, ":".join(laststep[1:]))

    def _match_iterable(self, iterable, key):
        """
        Searches for a key match within a given iterable.
        Raises ValueError if not found.
        """
        for obj in iterable:
            if self._matches(obj, key):
                return obj
        raise ValueError("Object named '%s' does not exist" % key)

    def _get_section_by_path(self, path):
        """
        Returns a Section by a given path.
        Raises ValueError if not found.
        """
        if path.startswith("/"):
            if len(path) == 1:
                raise ValueError("Not a valid path")
            doc = self.document
            if doc is not None:
                return doc._get_section_by_path(path[1:])
            raise ValueError(
                "A section with no Document cannot resolve absolute path")

        pathlist = path.split("/")
        if len(pathlist) > 1:
            if pathlist[0] == "..":
                found = self.parent
            elif pathlist[0] == ".":
                found = self
            else:
                found = self._match_iterable(self.sections, pathlist[0])

            if found:
                return found._get_section_by_path("/".join(pathlist[1:]))
            raise ValueError("Section named '%s' does not exist" % pathlist[0])
        else:
            return self._match_iterable(self.sections, pathlist[0])

    def find(self, key=None, type=None, findAll=False, include_subtype=False):
        """ Return the first subsection named *key* of type *type* """
        ret = []
        if type:
            type = type.lower()

        for s in self._sections:
            if self._matches(s, key, type, include_subtype=include_subtype):
                if findAll:
                    ret.append(s)
                else:
                    return s
        if ret:
            return ret

    def find_related(self, key=None, type=None, children=True, siblings=True,
                     parents=True, recursive=True, findAll=False):
        """
        Finds a related section named *key* and/or *type*

          * by searching its children’s children if *children* is True
            if *recursive* is true all leave nodes will be searched
          * by searching its siblings if *siblings* is True
          * by searching the parent element if *parents* is True
            if *recursive* is True all parent nodes until the root are searched
          * if *findAll* is True, returns a list of all matching objects
        """
        ret = []
        if type:
            type = type.lower()

        if children:
            for section in self._sections:
                if self._matches(section, key, type):
                    if findAll:
                        ret.append(section)
                    else:
                        return section

                if recursive:
                    obj = section.find_related(key, type, children,
                                               siblings=False, parents=False,
                                               recursive=recursive,
                                               findAll=findAll)
                    if obj is not None:
                        if findAll:
                            ret += obj
                        else:
                            return obj

        if siblings and self.parent is not None:
            obj = self.parent.find(key, type, findAll)
            if obj is not None:
                if findAll:
                    ret += obj
                else:
                    return obj

        if parents:
            obj = self
            while obj.parent is not None:
                obj = obj.parent
                if self._matches(obj, key, type):
                    if findAll:
                        ret.append(obj)
                    else:
                        return obj
                if not recursive:
                    break

        if ret:
            return ret
        return None

    def get_path(self):
        """
        Returns the absolute path of this section
        """
        node = self
        path = []
        while node.parent is not None:
            path.insert(0, node.name)
            node = node.parent
        return "/" + "/".join(path)

    @staticmethod
    def _get_relative_path(a, b):
        """
        Returns a relative path for navigation from dir *a* to dir *b*

        If the common parent of both is "/", return an absolute path
        """
        a += "/"
        b += "/"
        parent = posixpath.dirname(posixpath.commonprefix([a, b]))
        if parent == "/":
            return b[:-1]

        a = posixpath.relpath(a, parent)
        b = posixpath.relpath(b, parent)
        if a == ".":
            return b

        return posixpath.normpath("../" * (a.count("/") + 1) + b)

    def get_relative_path(self, section):
        """
        Returns a relative (file)path to point to section
        like (e.g. ../other_section)

        If the common parent of both sections is the document (i.e. /),
        return an absolute path
        """
        a = self.get_path()
        b = section.get_path()
        return self._get_relative_path(a, b)

    def clean(self):
        """
        Runs clean() on all immediate child-sections causing any resolved links
        or includes to be unresolved.

        This should be called for the document prior to saving.
        """
        for i in self:
            i.clean()

    def clone(self, children=True):
        """
        Clone this object recursively allowing to copy it independently
        to another document
        """
        from odml.section import BaseSection
        obj = super(Sectionable, self).clone(children)
        obj._parent = None
        obj._sections = SmartList(BaseSection)
        if children:
            for s in self._sections:
                obj.append(s.clone())

        return obj

    @property
    def repository(self):
        """ A URL to a terminology. """
        return self._repository

    @repository.setter
    def repository(self, url):
        if not url:
            url = None
        self._repository = url
        if url:
            terminology.deferred_load(url)

    def get_repository(self):
        """
        Return the current applicable repository (may be inherited from a
        parent) or None
        """
        return self._repository



"""
This file tests proper conversion of complex v1.0 odML files
to v1.1 with all supported file formats.
"""

import os
import shutil
import tempfile
import unittest

from odml import load
from odml.tools.version_converter import VersionConverter as VC


class TestVersionConverterIntegration(unittest.TestCase):
    def setUp(self):
        # Set up test environment
        dir_path = os.path.dirname(os.path.realpath(__file__))

        self.jsonfile = os.path.join(dir_path, "resources", "version_conversion_int.json")
        self.xmlfile = os.path.join(dir_path, "resources", "version_conversion_int.xml")
        self.yamlfile = os.path.join(dir_path, "resources", "version_conversion_int.yaml")

        self.tmp_dir = tempfile.mkdtemp(suffix=".odml")
        self.outfile = os.path.join(self.tmp_dir, "version_conversion.xml")

    def tearDown(self):
        if os.path.exists(self.tmp_dir):
            shutil.rmtree(self.tmp_dir)

    def test_convert_xml(self):
        VC(self.xmlfile).write_to_file(self.outfile)
        self.assertTrue(os.path.exists(self.outfile))

        self.check_result()

    def test_convert_yaml(self):
        with self.assertRaises(Exception):
            VC(self.yamlfile).write_to_file(self.outfile)

        VC(self.yamlfile).write_to_file(self.outfile, "YAML")
        self.assertTrue(os.path.exists(self.outfile))

        self.check_result()

    def test_convert_json(self):
        with self.assertRaises(Exception):
            VC(self.jsonfile).write_to_file(self.outfile)

        VC(self.jsonfile).write_to_file(self.outfile, "JSON")
        self.assertTrue(os.path.exists(self.outfile))

        self.check_result()

    def check_result(self):
        doc = load(self.outfile)

        # Test document attribute export
        self.assertEqual(doc.author, "author")
        self.assertEqual(doc.version, "v1.13")
        repo = "http://portal.g-node.org/odml/terminologies/v1.1/terminologies.xml"
        self.assertEqual(doc.repository, repo)
        self.assertEqual(len(doc.sections), 3)

        sec = doc.sections["sec_one"]
        self.assertEqual(sec.definition, "def s1")
        self.assertEqual(sec.reference, "ref s1")
        self.assertEqual(sec.type, "mainsec")
        self.assertEqual(len(sec.sections), 1)
        self.assertEqual(len(sec.properties), 3)

        prop = sec.properties["prop_one"]
        self.assertEqual(prop.definition, "def prop1")
        self.assertEqual(prop.dependency, "dep p1")
        self.assertEqual(prop.dependency_value, "dep val p1")
        self.assertEqual(prop.reference, "ref val 1")
        self.assertEqual(prop.uncertainty, "11")
        self.assertEqual(prop.unit, "arbitrary")
        self.assertEqual(prop.value_origin, "filename val 1")
        self.assertEqual(prop.dtype, "string")
        self.assertEqual(len(prop.value), 3)

        prop = sec.properties["prop_two"]
        self.assertEqual(len(prop.value), 8)

        prop = sec.properties["prop_three"]
        self.assertEqual(len(prop.value), 0)



import unittest
import sys
import odml

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


class TestTypes(unittest.TestCase):

    def setUp(self):
        # Capture the output printed by the functions to STDOUT, and use it for
        # testing purposes.
        self.captured_stdout = StringIO()
        sys.stdout = self.captured_stdout

        self.doc = odml.Document(author='Rave', version='1.0')
        s1 = odml.Section(name='Cell')
        p1 = odml.Property(name='Type', value='Rechargeable')
        s1.append(p1)

        s2 = odml.Section(name='Electrolyte')
        p2 = odml.Property(name='Composition', value='Ni-Cd')
        s2.append(p2)
        s1.append(s2)

        s3 = odml.Section(name='Electrode')
        p3 = odml.Property(name='Material', value='Nickel')
        p4 = odml.Property(name='Models', value=['AA', 'AAA'])
        s3.append(p3)
        s3.append(p4)
        s2.append(s3)

        self.doc.append(s1)

    def test_dump_doc(self):
        # This test dumps the whole document and checks it word by word.
        # If possible, maybe some better way of testing this ?
        odml.tools.dumper.dumpDoc(self.doc)
        output = [x.strip() for x in self.captured_stdout.getvalue().split('\n') if x]
        expected_output = []
        expected_output.append("*Cell ()")
        expected_output.append(":Type (value=Rechargeable, dtype='string')")
        expected_output.append("*Electrolyte ()")
        expected_output.append(":Composition (value=Ni-Cd, dtype='string')")
        expected_output.append("*Electrode ()")
        expected_output.append(":Material (value=Nickel, dtype='string')")
        expected_output.append(":Models (value=[AA,AAA], dtype='string')")
        self.assertEqual(len(output), len(expected_output))
        for i in range(len(output)):
            self.assertEqual(output[i], expected_output[i])

        # Discard the document output from stdout stream
        self.captured_stdout.seek(0)
        self.captured_stdout.truncate(0)



import unittest
import sys
from datetime import datetime as dt, date, time
from odml import Property, Section, Document
from odml.tools.xmlparser import XMLReader, XMLWriter


class TestInferType(unittest.TestCase):

    def test_string(self):
        p = Property("test", value="somestring")
        assert(p.dtype == "string")
        if sys.version_info < (3, 0):
            assert isinstance(p.value[0], unicode)
        else:
            assert isinstance(p.value[0], str)

    def test_text(self):
        p = Property("test", value="some\nstring")
        assert(p.dtype == "text")
        if sys.version_info < (3, 0):
            assert isinstance(p.value[0], unicode)
        else:
            assert isinstance(p.value[0], str)

    def test_int(self):
        p = Property("test", value=111)
        assert(p.dtype == "int")
        assert isinstance(p.value[0], int)

    def test_float(self):
        p = Property("test", value=3.14)
        assert(p.dtype == "float")
        assert isinstance(p.value[0], float)

    def test_datetime(self):
        p = Property("test", value=dt.now())
        assert(p.dtype == "datetime")
        assert isinstance(p.value[0], dt)

    def test_date(self):
        p = Property("test", dt.now().date())
        assert(p.dtype == "date")
        assert isinstance(p.value[0], date)

    def test_time(self):
        p = Property("test", value=dt.now().time())
        assert(p.dtype == "time")
        assert isinstance(p.value[0], time)

    def test_boolean(self):
        p = Property("test", True)
        assert(p.dtype == "boolean")
        assert isinstance(p.value[0], bool)

        p = Property("test", False)
        assert(p.dtype == "boolean")
        assert isinstance(p.value[0], bool)

    def test_read_write(self):
        doc = Document("author")
        sec = Section("name", "type")
        doc.append(sec)
        sec.append(Property("strprop", "somestring"))
        sec.append(Property("txtprop", "some\ntext"))
        sec.append(Property("intprop", 200))
        sec.append(Property("floatprop", 2.00))
        sec.append(Property("datetimeprop", dt.now()))
        sec.append(Property("dateprop", dt.now().date()))
        sec.append(Property("timeprop", dt.now().time()))
        sec.append(Property("boolprop", True))
        if sys.version_info < (3, 0):
            str_doc = unicode(XMLWriter(doc))
        else:
            str_doc = str(XMLWriter(doc))
        new_doc = XMLReader().from_string(str_doc)
        new_sec = new_doc.sections[0]

        p = new_sec.properties["strprop"]
        assert(p.dtype == "string")
        if sys.version_info < (3, 0):
            assert isinstance(p.value[0], unicode)
        else:
            assert isinstance(p.value[0], str)

        p = new_sec.properties["txtprop"]
        assert(p.dtype == "text")
        if sys.version_info < (3, 0):
            assert isinstance(p.value[0], unicode)
        else:
            assert isinstance(p.value[0], str)

        p = new_sec.properties["intprop"]
        assert(p.dtype == "int")
        assert isinstance(p.value[0], int)

        p = new_sec.properties["floatprop"]
        assert(p.dtype == "float")
        assert isinstance(p.value[0], float)

        p = new_sec.properties["datetimeprop"]
        assert(p.dtype == "datetime")
        assert isinstance(p.value[0], dt)

        p = new_sec.properties["dateprop"]
        assert(p.dtype == "date")
        assert isinstance(p.value[0], date)

        p = new_sec.properties["timeprop"]
        assert(p.dtype == "time")
        assert isinstance(p.value[0], time)

        p = new_sec.properties["boolprop"]
        assert(p.dtype == "boolean")
        assert isinstance(p.value[0], bool)



import unittest
from odml import Section


class TestFindSections(unittest.TestCase):

    def setUp(self):
        self.root_section = Section("Type test", "test")
        self.root_section.append(Section("sub_1", "sub_1"))
        self.root_section.append(Section("sub_2", "sub_2"))
        self.root_section.append(Section("sub_2_b", "sub_2"))
        self.root_section.append(Section("sub_3", "sub_2/sub_3"))

    def test_find_by_name(self):
        ret = self.root_section.find("sub_1")
        self.assertTrue(ret.name == "sub_1")

        ret = self.root_section.find("unknown_type")
        self.assertIsNone(ret)

    def test_find_by_type(self):
        ret = self.root_section.find(type="sub_1")
        self.assertTrue(ret is not None and ret.type == "sub_1")

        ret = self.root_section.find(type="sub_2", findAll=True)
        self.assertTrue(len(ret) == 2)

        ret = self.root_section.find(key=None, type="sub_2", findAll=True,
                                     include_subtype=True)

        self.assertTrue(len(ret) == 3)

    def test_find_by_name_and_type(self):
        ret = self.root_section.find(key="sub_1", type="sub_1")
        self.assertTrue(ret.name == "sub_1")

        ret = self.root_section.find(key="sub_1", type="sub_2")
        self.assertIsNone(ret)



"""
Dumps odML-Structures
"""
from .xmlparser import to_csv


def get_props(obj, props):
    out = []
    for p in props:
        if hasattr(obj, p):
            x = getattr(obj, p)
            if x is not None:
                if isinstance(x, list) or isinstance(x, tuple):
                    out.append("%s=%s" % (p, to_csv(x)))
                else:
                    out.append("%s=%s" % (p, repr(x)))

    return ", ".join(out)


def dumpProperty(property, indent=1):
    # TODO : (PEP8) Find a better way to split the following line
    print("%*s:%s (%s)" % (indent, " ", property.name,
          get_props(property, ["definition", "value", "uncertainty", "unit",
                               "dtype", "value_reference", "dependency",
                               "dependencyValue"])))


def dumpSection(section, indent=1):
    if section is None:
        return

    # TODO : (PEP8) Find a better way to split the following line
    print("%*s*%s (%s)" % (indent, " ", section.name,
          get_props(section, ["type", "definition", "link",
                              "include", "repository"])))

    for prop in section.properties:
        dumpProperty(prop, indent + 1)

    for sub in section.sections:
        dumpSection(sub, indent * 2)


def dumpDoc(doc):
    for sec in doc:
        dumpSection(sec)



import re
from abc import ABCMeta, abstractmethod

from rdflib import Namespace, RDF
from rdflib.plugins.sparql import prepareQuery

from ..format import Document
from ..format import Property
from ..format import Section


class BaseQueryCreator:

    __metaclass__ = ABCMeta

    possible_query_variables = {'d': 'Document', 's': 'Section',
                                'p': 'Property', 'v': 'Bag URI', 'value': 'Value'}

    possible_q_dict_keys = ['Doc', 'Sec', 'Prop']

    def __init__(self, q_dict=None):
        """
        :param q_dict: dictionary with query parameters
        """
        self.q_dict = q_dict if q_dict else {}
        self.query = ''
        super(BaseQueryCreator, self).__init__()

    @abstractmethod
    def get_query(self, q_str, q_parser):
        pass
    
    @abstractmethod
    def _prepare_query(self):
        pass


class BaseQueryParser:

    __metaclass__ = ABCMeta

    def __init__(self):
        self.q_dict = {}

    @abstractmethod
    def parse_query_string(self, q_str):
        pass


class QueryParserFuzzy(BaseQueryParser):

    def __init__(self):
        super(QueryParserFuzzy, self).__init__()

    def parse_query_string(self, q_str):
        """
        Parse query string and returns dict object with parameters.
        :param q_str: query string.
                      Example: FIND sec(name, type) prop(type) HAVING Stimulus, Contrast
        :return: dict object.
                 Example: {'Sec': ['name', 'type'],
                           'Doc': ['author'],
                           'Search': ['Stimulus', 'Contrast']}
        """
        self.q_dict = {}
        find_pattern = re.compile("FIND(.*?)HAVING")
        find_group = re.search(find_pattern, q_str).group(1).strip()
        if find_group:
            self._parse_find(find_group)

        having_pattern = re.compile("HAVING(.*)")
        having_group = re.search(having_pattern, q_str).group(1).strip()
        if having_group:
            if 'Search' in self.q_dict.keys():
                raise ValueError('Search values are already parsed')
            self._parse_having(having_group)
        else:
            raise ValueError('Search values in having part were not specified')

        return self.q_dict

    def _parse_find(self, find_part):
        """
        Parses find string part into list of specific keys to whih search values would be apllied
        e.g. 'sec(name, type) prop(name)' into {'Sec': ['name', 'type'], 'Prop': ['name']} .
        
        :param find_part: string which represent list of searchable odML data model objects 
                            like document(doc), sections(sec) or properties(prop).
                            e.g. 'sec(name, type) prop(name)'
        """
        doc_pattern = re.compile("(doc|document)\(.*?\)")
        doc = re.search(doc_pattern, find_part)
        if doc:
            self._parse_doc(doc)

        sec_pattern = re.compile("(sec|section)\(.*?\)")
        sec = re.search(sec_pattern, find_part)
        if sec:
            self._parse_sec(sec)

        prop_pattern = re.compile("(prop|property)\(.*?\)")
        prop = re.search(prop_pattern, find_part)
        if prop:
            self._parse_prop(prop)

    def _parse_doc(self, doc):
        p = re.compile("[\(|, ](id|author|date|version|repository|sections)[\)|,]")
        if doc:
            self.q_dict['Doc'] = re.findall(p, doc.group(0))

    def _parse_sec(self, sec):
        p = re.compile("[\(|, ](id|name|definition|type|repository|reference|sections|properties)[\)|,]")
        if sec:
            self.q_dict['Sec'] = re.findall(p, sec.group(0))

    def _parse_prop(self, prop):
        p = re.compile("[\(|, ](id|name|definition|dtype|unit|uncertainty|reference|value_origin)[\)|,]")
        if prop:
            self.q_dict['Prop'] = re.findall(p, prop.group(0))

    def _parse_having(self, having_part):
        """
        Parses search value string into list of specific values 
        e.g. 'Stimulus, Contrast, Date' into list [Stimulus, Contrast, Date].
        
        :param having_part: string with search values, e.g. 'Stimulus, Contrast'
                      Also spaces errors in the string like 'Stimulus,    ,  Contrast' will be ignored.
        """
        search_values_list = []
        search_params = re.compile("(.*?)(?:,|$)")
        if having_part:
            search_values = re.findall(search_params, having_part)
            for v in search_values:
                if v.strip():
                    search_values_list.append(v.strip())
        self.q_dict['Search'] = search_values_list


class QueryParser(BaseQueryParser):

    def __init__(self):
        super(QueryParser, self).__init__()

    def parse_query_string(self, q_str):
        """
        :param q_str: query string
                      Example: doc(author:D. N. Adams) section(name:Stimulus) prop(name:Contrast, value:20, unit:%)
        :return: dict object
                 Example: {'Sec': [('name', 'Stimulus')],
                           'Doc': [('author', 'D. N. Adams')],
                           'Prop': [('name', 'Contrast'), ('value':[20]), ('unit':'%')]}
        """
        doc_pattern = re.compile("(doc|document)\(.*?\)")
        doc = re.search(doc_pattern, q_str)
        if doc:
            self._parse_doc(doc)

        sec_pattern = re.compile("(sec|section)\(.*?\)")
        sec = re.search(sec_pattern, q_str)
        if sec:
            self._parse_sec(sec)

        prop_pattern = re.compile("(prop|property)\(.*?\)")
        prop = re.search(prop_pattern, q_str)
        if prop:
            self._parse_prop(prop)
        
        return self.q_dict

    def _parse_doc(self, doc):
        p = re.compile("[, |\(](id|author|date|version|repository|sections):(.*?)[,|\)]")
        if doc:
            self.q_dict['Doc'] = re.findall(p, doc.group(0))

    def _parse_sec(self, sec):
        p = re.compile("[, |\(](id|name|definition|type|repository|reference|sections|properties):(.*?)[,|\)]")
        if sec:
            self.q_dict['Sec'] = re.findall(p, sec.group(0))

    def _parse_prop(self, prop):
        p = re.compile("[, |\(](id|name|definition|dtype|unit|uncertainty|reference|value_origin):(.*?)[,|\)]")
        if prop:
            self.q_dict['Prop'] = re.findall(p, prop.group(0))

            p_value = re.compile("value:\[(.*)]")

            value_group = re.findall(p_value, prop.group(0))
            if value_group:
                values = re.split(", ?", value_group[0])
                self.q_dict['Prop'].append(('value', values))


class QueryCreator(BaseQueryCreator):
    """ 
    Class for simplifying the creation of prepared SPARQL queries 
    
    Usage:
        q = "doc(author:D. N. Adams) section(name:Stimulus) prop(name:Contrast, value:20, unit:%)"
        prepared_query = QueryCreator().get_query(q, QueryParser())
        
        q = "FIND sec(name, type) prop(name) HAVING Recording, Recording-2012-04-04-ab, Date"
        prepared_query = QueryCreator().get_query(q, QueryParser2())
    """

    def __init__(self, q_dict=None):
        """
        :param q_dict: dictionary with query parameters
        """
        super(QueryCreator, self).__init__(q_dict)

    def get_query(self, q_str=None, q_parser=None):
        """
        :param q_parser: one of possible query parsers.
        :param q_str: doc(author:D. N. Adams) section(name:Stimulus) prop(name:Contrast, value:20, unit:%)
        :return rdflib prepare query.
        """
        # TODO find out if the validation for the q_str is important
        # We can possibly warn about not used parts and print the parsed dictionary
        if not self.q_dict:
            if not q_str:
                raise AttributeError("Please fulfill q_str param (query string)")
            elif not q_parser:
                raise AttributeError("Please fulfill q_parser param (query parser)")
            self.q_dict = q_parser.parse_query_string(q_str)
        self._prepare_query()
        return prepareQuery(self.query, initNs={"odml": Namespace("https://g-node.org/projects/odml-rdf#"),
                                                "rdf": RDF})

    def _prepare_query(self):
        """
        Creates rdflib query using parameters from self.q_dict.
        :return: string representing rdflib query.
        """

        odml_uri = "https://g-node.org/projects/odml-rdf#"
        self.query = 'SELECT * WHERE {\n'

        if 'Doc' in self.q_dict.keys():
            doc_attrs = self.q_dict['Doc']
            if len(doc_attrs) > 0:
                self.query += '?d rdf:type odml:Document .\n'
                for i in doc_attrs:
                    if len(i) > 2:
                        raise ValueError("Attributes in the query \"{}\" are not valid.".format(i))
                    else:
                        attr = Document.rdf_map(i[0])
                        if attr:
                            self.query += '?d {0} \"{1}\" .\n'.format(re.sub(odml_uri,
                                                                             "odml:", attr), i[1])
        if 'Sec' in self.q_dict.keys():
            sec_attrs = self.q_dict['Sec']
            if len(sec_attrs) > 0:
                self.query += '?d odml:hasSection ?s .\n' \
                              '?s rdf:type odml:Section .\n'
                for i in sec_attrs:
                    if len(i) > 2:
                        raise ValueError("Attributes in the query \"{}\" are not valid.".format(i))
                    else:
                        attr = Section.rdf_map(i[0])
                        if attr:
                            self.query += '?s {0} \"{1}\" .\n'.format(re.sub(odml_uri,
                                                                             "odml:", attr), i[1])
        if 'Prop' in self.q_dict.keys():
            prop_attrs = self.q_dict['Prop']
            if len(prop_attrs) > 0:
                self.query += '?s odml:hasProperty ?p .\n' \
                              '?p rdf:type odml:Property .\n'
                for i in prop_attrs:
                    if len(i) > 2:
                        raise ValueError("Attributes in the query \"{}\" are not valid.".format(i))
                    elif i[0] == 'value':
                        values = i[1]
                        if values:
                            self.query += "?p odml:hasValue ?v .\n?v rdf:type rdf:Bag .\n"
                            for v in values:
                                self.query += '?v rdf:li \"{}\" .\n'.format(v)
                    else:
                        attr = Property.rdf_map(i[0])
                        if attr:
                            self.query += '?p {0} \"{1}\" .\n'.format(re.sub(odml_uri,
                                                                             "odml:", attr), i[1])

        self.query += '}\n'
        return self.query



import argparse
import os
import re
import sys

import odml
from .rdf_converter import RDFWriter
from .version_converter import VersionConverter

try:
    unicode = unicode
except NameError:
    unicode = str


class FormatConverter(object):

    _conversion_formats = {
        'v1_1': '.xml',
        'odml': '.odml',
        # rdflib version "4.2.2" serialization formats
        'xml': '.rdf',
        'pretty-xml': '.rdf',
        'trix': '.rdf',
        'n3': '.n3',
        'turtle': '.ttl',
        'ttl': '.ttl',
        'ntriples': '.nt',
        'nt': '.nt',
        'nt11': '.nt',
        'trig': '.trig',
        'json-ld': '.jsonld'
    }

    @classmethod
    def convert(cls, args=None):
        """
        Enable usage of the argparse for calling convert_dir(...)
        Example:
            1) >> python format_converter.py ./..path../input_dir v1_1 -out ./..path../output_dir -r
            
               Convert files from the path <./..path../input_dir> to .xml odml version 1.1,
               writes them into <./..path../output_dir> including subdirectories 
               and its files from the input path.
            
            2) >> python format_converter.py ./..path../input_dir odml
            
               Converts files from path <./..path../input_dir> to .odml,
               writes them into <./..path../input_dir_odml> not including subdirectories.
        """
        parser = argparse.ArgumentParser(description="Convert directory with odml files to another format")
        parser.add_argument("input_dir", help="Path to input directory")
        parser.add_argument("result_format", choices=list(cls._conversion_formats),
                            help="Format of output files")
        parser.add_argument("-out", "--output_dir", help="Path for output directory")
        parser.add_argument("-r", "--recursive", action="store_true",
                            help="Enable converting files from subdirectories")
        args = parser.parse_args(args)
        r = True if args.recursive else False
        cls.convert_dir(args.input_dir, args.output_dir, r, args.result_format)

    @classmethod
    def convert_dir(cls, input_dir, output_dir, parse_subdirs, res_format):
        """
        Convert files from given input directory to the specified res_format.
        :param input_dir: Path to input directory
        :param output_dir: Path for output directory. If None, new directory will be created on the same level as input
        :param parse_subdirs: If True enable converting files from subdirectories
        :param res_format: Format of output files. 
                           Possible choices: "v1_1" (converts to version 1.1 from version 1 xml)
                                             "odml" (converts to .odml from version 1.1 .xml files)
                                             "turtle", "nt" etc. (converts to rdf formats from version 1.1 .odml files)
                                             (see full list of rdf serializers in FormatConverter._conversion_formats)
        """
        if res_format not in cls._conversion_formats:
            raise ValueError("Format for output files is incorrect. "
                             "Please choose from the list: {}".format(cls._conversion_formats.keys()))

        cls._check_input_output_directory(input_dir, output_dir)
        input_dir = os.path.join(input_dir, '')

        if output_dir is None:
            input_dir_name = os.path.basename(os.path.dirname(input_dir))
            root_dir = os.path.dirname(os.path.dirname(input_dir))  # find the directory that contains input_dir
            output_dir_name = input_dir_name + "_" + res_format
            output_dir = os.path.join(root_dir, output_dir_name)
            cls._create_sub_directory(output_dir)

        output_dir = os.path.join(output_dir, '')

        if not parse_subdirs:
            for file_name in os.listdir(input_dir):
                if os.path.isfile(os.path.join(input_dir, file_name)):
                    cls._convert_file(os.path.join(input_dir, file_name), os.path.join(output_dir, file_name),
                                      res_format)
        else:
            for dir_path, dir_names, file_names in os.walk(input_dir):
                for file_name in file_names:
                    in_file_path = os.path.join(dir_path, file_name)
                    out_dir = re.sub(r"" + input_dir, r"" + output_dir, dir_path)
                    out_file_path = os.path.join(out_dir, file_name)
                    cls._create_sub_directory(out_dir)
                    cls._convert_file(in_file_path, out_file_path, res_format)

    @classmethod
    def _convert_file(cls, input_path, output_path, res_format):
        """
        Convert a file from given input_path to res_format.
        """
        if res_format == "v1_1":
            VersionConverter(input_path).write_to_file(output_path)
        elif res_format == "odml":
            if not output_path.endswith(".odml"):
                p, _ = os.path.splitext(output_path)
                output_path = p + ".odml"
            odml.save(odml.load(input_path), output_path)
        elif res_format in cls._conversion_formats:
            if not output_path.endswith(cls._conversion_formats[res_format]):
                p, _ = os.path.splitext(output_path)
                output_path = p + cls._conversion_formats[res_format]
            RDFWriter(odml.load(input_path)).write_file(output_path, res_format)

    @staticmethod
    def _create_sub_directory(dir_path):
        """
        Creates the new directory to store the converted file.
        """
        if not os.path.isdir(dir_path):
            os.makedirs(dir_path)

    @staticmethod
    def _check_input_output_directory(input_dir, output_dir):
        """
        Checks if provided directory is valid - not None, is directory and not a root folder in the File System 
        if output dir was not provided.
        Raise relevant exceptions.
        """
        if not input_dir or not os.path.isdir(input_dir):
            raise ValueError("The path to input directory is not a valid path")

        if output_dir is not None and not os.path.isdir(output_dir):
            raise ValueError("The path to output directory is not a valid path")

        if not output_dir:
            if os.path.dirname(input_dir) == input_dir:
                raise ValueError("The input directory cannot be a root folder of the File System if "
                                 "output directory was not specified")

if __name__ == "__main__":
    FormatConverter.convert(sys.argv[1:])



"""
This file tests proper creation, saving and loading
of odML Sections with all supported odML parsers.
"""

import os
import shutil
import tempfile
import unittest

import odml


class TestSectionIntegration(unittest.TestCase):

    def setUp(self):
        # Set up test environment
        self.tmp_dir = tempfile.mkdtemp(suffix=".odml")

        self.json_file = os.path.join(self.tmp_dir, "test.json")
        self.xml_file = os.path.join(self.tmp_dir, "test.xml")
        self.yaml_file = os.path.join(self.tmp_dir, "test.yaml")

        # Set up odML document stub
        doc = odml.Document()
        self.doc = doc

    def tearDown(self):
        if os.path.exists(self.tmp_dir):
            shutil.rmtree(self.tmp_dir)

    def save_load(self):
        """
        Helper method to save and load the current state of the document
        with all supported parsers.
        :return: jdoc ... document loaded from JSON file
                 xdoc ... document loaded from XML file
                 ydoc ... document loaded from YAML file
        """
        odml.save(self.doc, self.json_file, "JSON")
        jdoc = odml.load(self.json_file, "JSON")

        odml.save(self.doc, self.xml_file)
        xdoc = odml.load(self.xml_file)

        odml.save(self.doc, self.yaml_file, "YAML")
        ydoc = odml.load(self.yaml_file, "YAML")

        return jdoc, xdoc, ydoc

    def test_id(self):
        # Test correct save and load of generated id.
        sec_name = "empty_id"
        sec = odml.Section(name=sec_name, parent=self.doc)

        jdoc, xdoc, ydoc = self.save_load()

        self.assertEqual(jdoc.sections[sec_name].oid, sec.oid)
        self.assertEqual(xdoc.sections[sec_name].oid, sec.oid)
        self.assertEqual(ydoc.sections[sec_name].oid, sec.oid)

        # Test correct save and load of assigned id.
        sec_name = "assigned_id"
        assigned_id = "79b613eb-a256-46bf-84f6-207df465b8f7"
        _ = odml.Section(name=sec_name, oid=assigned_id, parent=self.doc)

        jdoc, xdoc, ydoc = self.save_load()

        self.assertEqual(jdoc.sections[sec_name].oid, assigned_id)
        self.assertEqual(xdoc.sections[sec_name].oid, assigned_id)
        self.assertEqual(ydoc.sections[sec_name].oid, assigned_id)

    def test_simple_attributes(self):
        """
        This test checks correct writing and loading of 'simple'
        Section format attributes meaning all attributes that
        do not require any special handling when they are set.
        """
        s_name = "section"
        s_type = "mellow"
        s_repo = "unresolvable"
        s_ref = "4 8 15 16 23 42"
        s_def = "undefined"

        _ = odml.Section(name=s_name, type=s_type, repository=s_repo,
                         reference=s_ref, definition=s_def, parent=self.doc)

        jdoc, xdoc, ydoc = self.save_load()

        # Test correct JSON save and load.
        jsec = jdoc.sections[s_name]
        self.assertEqual(jsec.name, s_name)
        self.assertEqual(jsec.type, s_type)
        self.assertEqual(jsec.repository, s_repo)
        self.assertEqual(jsec.reference, s_ref)
        self.assertEqual(jsec.definition, s_def)

        # Test correct XML save and load.
        xsec = xdoc.sections[s_name]
        self.assertEqual(xsec.name, s_name)
        self.assertEqual(xsec.type, s_type)
        self.assertEqual(xsec.repository, s_repo)
        self.assertEqual(xsec.reference, s_ref)
        self.assertEqual(xsec.definition, s_def)

        # Test correct YAML save and load.
        ysec = ydoc.sections[s_name]
        self.assertEqual(ysec.name, s_name)
        self.assertEqual(ysec.type, s_type)
        self.assertEqual(ysec.repository, s_repo)
        self.assertEqual(ysec.reference, s_ref)
        self.assertEqual(ysec.definition, s_def)

    def test_children(self):
        """
        This test checks correct writing and loading of Section and Property
        children of a Section.
        """
        root = odml.Section(name="root", parent=self.doc)

        # Lvl 1 child Sections
        sec_lvl_11 = odml.Section(name="sec_11", parent=root)
        _ = odml.Section(name="sec_12", parent=root)

        # Lvl 1 child Properties
        _ = odml.Property(name="prop_11", parent=root)
        _ = odml.Property(name="prop_12", parent=root)

        # Lvl 2 child Sections
        sec_lvl_21 = odml.Section(name="sec_21", parent=sec_lvl_11)
        _ = odml.Section(name="sec_22", parent=sec_lvl_11)
        _ = odml.Section(name="sec_23", parent=sec_lvl_11)

        # Lvl 2 child Properties
        _ = odml.Property(name="prop_21", parent=sec_lvl_11)
        _ = odml.Property(name="prop_22", parent=sec_lvl_11)
        _ = odml.Property(name="prop_23", parent=sec_lvl_11)

        # Lvl 3 child Sections
        _ = odml.Section(name="sec_31", parent=sec_lvl_21)
        _ = odml.Section(name="sec_32", parent=sec_lvl_21)
        _ = odml.Section(name="sec_33", parent=sec_lvl_21)
        _ = odml.Section(name="sec_34", parent=sec_lvl_21)

        # Lvl 3 child Properties
        _ = odml.Property(name="prop_31", parent=sec_lvl_21)
        _ = odml.Property(name="prop_32", parent=sec_lvl_21)
        _ = odml.Property(name="prop_33", parent=sec_lvl_21)
        _ = odml.Property(name="prop_34", parent=sec_lvl_21)

        jdoc, xdoc, ydoc = self.save_load()

        # Test correct JSON save and load.
        jsec = jdoc.sections[root.name]
        self.assertEqual(len(jsec.sections), 2)
        self.assertEqual(len(jsec.properties), 2)

        jsec_lvl_1 = jsec[sec_lvl_11.name]
        self.assertEqual(len(jsec_lvl_1.sections), 3)
        self.assertEqual(len(jsec_lvl_1.properties), 3)

        jsec_lvl_2 = jsec_lvl_1[sec_lvl_21.name]
        self.assertEqual(len(jsec_lvl_2.sections), 4)
        self.assertEqual(len(jsec_lvl_2.properties), 4)

        # Test correct XML save and load.
        xsec = xdoc.sections[root.name]
        self.assertEqual(len(xsec.sections), 2)
        self.assertEqual(len(xsec.properties), 2)

        xsec_lvl_1 = xsec[sec_lvl_11.name]
        self.assertEqual(len(xsec_lvl_1.sections), 3)
        self.assertEqual(len(xsec_lvl_1.properties), 3)

        xsec_lvl_2 = xsec_lvl_1[sec_lvl_21.name]
        self.assertEqual(len(xsec_lvl_2.sections), 4)
        self.assertEqual(len(xsec_lvl_2.properties), 4)

        # Test correct YAML save and load.
        ysec = ydoc.sections[root.name]
        self.assertEqual(len(ysec.sections), 2)
        self.assertEqual(len(ysec.properties), 2)

        ysec_lvl_1 = ysec[sec_lvl_11.name]
        self.assertEqual(len(ysec_lvl_1.sections), 3)
        self.assertEqual(len(ysec_lvl_1.properties), 3)

        ysec_lvl_2 = ysec_lvl_1[sec_lvl_21.name]
        self.assertEqual(len(ysec_lvl_2.sections), 4)
        self.assertEqual(len(ysec_lvl_2.properties), 4)

    def test_link(self):
        pass

    def test_include(self):
        pass



import datetime
import unittest

import odml.dtypes as typ


class TestTypes(unittest.TestCase):

    def setUp(self):
        pass

    def test_valid_type(self):
        # Test None
        self.assertTrue(typ.valid_type(None))

        # Test that all DTypes classify as valid dtypes.
        for curr_type in typ.DType:
            self.assertTrue(typ.valid_type(curr_type), "Invalid DType %s" % curr_type)

        # Test that provided shorthand dtypes return as valid dtypes.
        for curr_shorthand in typ._dtype_map.keys():
            self.assertTrue(typ.valid_type(curr_shorthand),
                            "Invalid dtype shorthand %s" % curr_shorthand)

        # Test valid tuple dtype
        self.assertTrue(typ.valid_type("2-tuple"))
        self.assertTrue(typ.valid_type("293939-tuple"))

        # Test invalid dtypes
        self.assertFalse(typ.valid_type(1))
        self.assertFalse(typ.valid_type("unsupported"))
        self.assertFalse(typ.valid_type("x-tuple"))

    def test_date(self):
        self.assertIsInstance(typ.date_get(None), datetime.date)
        self.assertIsInstance(typ.date_get(""), datetime.date)

        re = "^[0-9]{4}-(0[1-9]|1[0-2])-([0-2][0-9]|3[0-1])$"
        self.assertRegexpMatches(typ.date_get(None).strftime(typ.FORMAT_DATE), re)
        self.assertRegexpMatches(typ.date_get("").strftime(typ.FORMAT_DATE), re)

        date = datetime.date(2011, 12, 1)
        date_string = '2011-12-01'
        self.assertEqual(date, typ.date_get(date_string))
        self.assertEqual(date, typ.date_get(date))

        with self.assertRaises(TypeError):
            _ = typ.date_get([])
        with self.assertRaises(TypeError):
            _ = typ.date_get({})
        with self.assertRaises(TypeError):
            _ = typ.date_get(False)

        # Test fail on datetime.datetime
        with self.assertRaises(ValueError):
            _ = typ.date_get(datetime.datetime.now())

        # Test fail on datetime.time
        with self.assertRaises(TypeError):
            _ = typ.date_get(datetime.datetime.now().time())

        # Test fail on invalid string format
        with self.assertRaises(ValueError):
            _ = typ.date_get("11.11.1111")

    def test_time(self):
        self.assertIsInstance(typ.time_get(None), datetime.time)
        self.assertIsInstance(typ.time_get(""), datetime.time)

        re = "^[0-5][0-9]:[0-5][0-9]:[0-5][0-9]$"
        self.assertRegexpMatches(typ.time_get(None).strftime(typ.FORMAT_TIME), re)
        self.assertRegexpMatches(typ.time_get("").strftime(typ.FORMAT_TIME), re)

        time = datetime.time(12, 34, 56)
        time_string = '12:34:56'
        self.assertEqual(time, typ.time_get(time_string))
        self.assertEqual(time, typ.time_get(time))

        with self.assertRaises(TypeError):
            _ = typ.time_get([])
        with self.assertRaises(TypeError):
            _ = typ.time_get({})
        with self.assertRaises(TypeError):
            _ = typ.time_get(False)

        # Test fail on datetime.datetime
        with self.assertRaises(TypeError):
            _ = typ.time_get(datetime.datetime.now())

        # Test fail on datetime.date
        with self.assertRaises(TypeError):
            _ = typ.time_get(datetime.datetime.now().date())

        # Test fail on invalid string format
        with self.assertRaises(ValueError):
            _ = typ.time_get("11-11-11")

    def test_datetime(self):
        self.assertIsInstance(typ.datetime_get(None), datetime.datetime)
        self.assertIsInstance(typ.datetime_get(""), datetime.datetime)

        re = "^[0-9]{4}-(0[1-9]|1[0-2])-([0-2][0-9]|3[0-1]) " \
             "[0-5][0-9]:[0-5][0-9]:[0-5][0-9]$"
        self.assertRegexpMatches(typ.datetime_get(None).strftime(typ.FORMAT_DATETIME), re)
        self.assertRegexpMatches(typ.datetime_get("").strftime(typ.FORMAT_DATETIME), re)

        date = datetime.datetime(2011, 12, 1, 12, 34, 56)
        date_string = '2011-12-01 12:34:56'
        self.assertEqual(date, typ.datetime_get(date_string))
        self.assertEqual(date, typ.datetime_get(date))

        with self.assertRaises(TypeError):
            _ = typ.datetime_get([])
        with self.assertRaises(TypeError):
            _ = typ.datetime_get({})
        with self.assertRaises(TypeError):
            _ = typ.datetime_get(False)

        # Test fail on datetime.time
        with self.assertRaises(TypeError):
            _ = typ.datetime_get(datetime.datetime.now().time())

        # Test fail on datetime.date
        with self.assertRaises(TypeError):
            _ = typ.datetime_get(datetime.datetime.now().date())

        # Test fail on invalid string format
        with self.assertRaises(ValueError):
            _ = typ.datetime_get("11.11.1111 12:12:12")

    def test_int(self):
        self.assertEqual(typ.default_values("int"), typ.int_get(None))
        self.assertEqual(typ.default_values("int"), typ.int_get(""))

        self.assertIsInstance(typ.int_get(11), int)
        self.assertIsInstance(typ.int_get(1.1), int)
        self.assertIsInstance(typ.int_get("11"), int)
        self.assertEqual(typ.int_get("123456789012345678901"), 123456789012345678901)
        self.assertEqual(typ.int_get("-123456789012345678901"), -123456789012345678901)
        self.assertEqual(typ.int_get("123.45"), 123)

        with self.assertRaises(TypeError):
            _ = typ.int_get([])
        with self.assertRaises(TypeError):
            _ = typ.int_get({})
        with self.assertRaises(ValueError):
            _ = typ.int_get("fail")

    def test_float(self):
        self.assertEqual(typ.default_values("float"), typ.float_get(None))
        self.assertEqual(typ.default_values("float"), typ.float_get(""))

        self.assertIsInstance(typ.float_get(1), float)
        self.assertIsInstance(typ.float_get("1.1"), float)
        self.assertEqual(typ.float_get(123.45), 123.45)

        with self.assertRaises(TypeError):
            _ = typ.float_get([])
        with self.assertRaises(TypeError):
            _ = typ.float_get({})
        with self.assertRaises(ValueError):
            _ = typ.float_get("fail")

    def test_str(self):
        self.assertEqual(typ.default_values("string"), typ.str_get(None))
        self.assertEqual(typ.default_values("string"), typ.str_get(""))
        self.assertEqual(typ.default_values("string"), typ.str_get([]))
        self.assertEqual(typ.default_values("string"), typ.str_get({}))

        # Make sure boolean values are properly converted to string.
        self.assertEqual(typ.str_get(False), 'False')
        self.assertEqual(typ.str_get(True), 'True')

    def test_bool(self):
        self.assertEqual(typ.default_values("boolean"), typ.boolean_get(None))
        self.assertEqual(typ.default_values("boolean"), typ.boolean_get(""))
        self.assertEqual(typ.default_values("boolean"), typ.boolean_get([]))
        self.assertEqual(typ.default_values("boolean"), typ.boolean_get({}))

        true_values = [True, "TRUE", "true", "T", "t", "1", 1]
        for val in true_values:
            self.assertTrue(typ.boolean_get(val))

        false_values = [False, "FALSE", "false", "F", "f", "0", 0]
        for val in false_values:
            self.assertFalse(typ.boolean_get(val))

        with self.assertRaises(ValueError):
            typ.boolean_get("text")
        with self.assertRaises(ValueError):
            typ.boolean_get(12)
        with self.assertRaises(ValueError):
            typ.boolean_get(2.1)

    def test_tuple(self):
        self.assertIs(typ.tuple_get(""), None)
        self.assertIs(typ.tuple_get(None), None)

        self.assertEqual(typ.tuple_get("(39.12; 67.19)"), ["39.12", "67.19"])

        # Test fail on missing parenthesis.
        with self.assertRaises(AssertionError):
            _ = typ.tuple_get("fail")
        # Test fail on mismatching element count and count number.
        with self.assertRaises(AssertionError):
            _ = typ.tuple_get("(1; 2; 3)", 2)

    def test_dtype_none(self):
        self.assertEqual(typ.get({'name': 'Marie'}), "{'name': 'Marie'}")



"""
Dict parser converts the content of a dictionary
into a proper and verified odML document.
"""

from .. import format as odmlfmt
from ..info import FORMAT_VERSION
from .parser_utils import InvalidVersionException, ParserException


class DictWriter:
    """
    A writer to parse an odML document to a Python dictionary object equivalent.
    """

    def __init__(self):
        self.doc = None  # odML document

    def to_dict(self, odml_document):
        self.doc = odml_document
        parsed_doc = {}

        for i in odmlfmt.Document.arguments_keys:
            attr = i
            if i in odmlfmt.Document.map_keys:
                attr = odmlfmt.Document.map(i)

            if hasattr(odml_document, attr):
                if attr == 'sections':
                    sections = self.get_sections(odml_document.sections)
                    parsed_doc[attr] = sections
                else:
                    tag = getattr(odml_document, attr)

                    if tag:
                        # Always use the arguments key attribute name when saving
                        parsed_doc[i] = tag

        return parsed_doc

    def get_sections(self, section_list):
        section_seq = []

        for section in section_list:
            section_dict = {}

            for i in odmlfmt.Section.arguments_keys:
                attr = i

                if i in odmlfmt.Section.map_keys:
                    attr = odmlfmt.Section.map(i)

                if hasattr(section, attr):
                    if attr == 'properties':
                        properties = self.get_properties(section.properties)
                        section_dict[attr] = properties
                    elif attr == 'sections':
                        sections = self.get_sections(section.sections)
                        section_dict[attr] = sections
                    else:
                        tag = getattr(section, attr)

                        if tag:
                            # Always use the arguments key attribute name when saving
                            section_dict[i] = tag

            section_seq.append(section_dict)

        return section_seq

    @staticmethod
    def get_properties(props_list):
        props_seq = []

        for prop in props_list:
            prop_dict = {}

            for i in odmlfmt.Property.arguments_keys:
                attr = i
                if i in odmlfmt.Property.map_keys:
                    attr = odmlfmt.Property.map(i)

                if hasattr(prop, attr):
                    tag = getattr(prop, attr)
                    if isinstance(tag, tuple):
                        prop_dict[attr] = list(tag)
                    elif (tag == []) or tag:  # Even if 'value' is empty, allow '[]'
                        # Custom odML tuples require special handling
                        # for save loading from file.
                        if attr == "value" and prop.dtype and \
                                prop.dtype.endswith("-tuple") and len(prop.value) > 0:
                            prop_dict["value"] = "(%s)" % ";".join(prop.value[0])
                        else:
                            # Always use the arguments key attribute name when saving
                            prop_dict[i] = tag

            props_seq.append(prop_dict)

        return props_seq


class DictReader:
    """
    A reader to parse dictionaries with odML content into a proper odML document.
    """

    def __init__(self):
        self.parsed_doc = None  # Python dictionary object equivalent
        self.warnings = []

    def is_valid_attribute(self, attr, fmt):
        if attr in fmt.arguments_keys:
            return attr

        if fmt.revmap(attr):
            return attr

        msg = "Invalid element <%s> inside <%s> tag" % (attr, fmt.__class__.__name__)
        print(msg)
        self.warnings.append(msg)
        return None

    def to_odml(self, parsed_doc):
        self.parsed_doc = parsed_doc

        # Parse only odML documents of supported format versions.
        if 'Document' not in self.parsed_doc:
            msg = "Missing root element 'Document'"
            raise ParserException(msg)
        elif 'odml-version' not in self.parsed_doc:
            raise ParserException("Invalid odML document: Could not find odml-version.")

        elif self.parsed_doc.get('odml-version') != FORMAT_VERSION:
            msg = ("Cannot parse odML document with format version '%s'. \n"
                   "\tUse the 'tools.VersionConverter' to import previous odML formats."
                   % self.parsed_doc.get('odml-version'))
            raise InvalidVersionException(msg)

        self.parsed_doc = self.parsed_doc['Document']

        doc_attrs = {}
        doc_secs = []

        for i in self.parsed_doc:
            attr = self.is_valid_attribute(i, odmlfmt.Document)
            if attr == 'sections':
                doc_secs = self.parse_sections(self.parsed_doc['sections'])
            elif attr:
                # Make sure to always use the correct odml format attribute name
                doc_attrs[odmlfmt.Document.map(attr)] = self.parsed_doc[i]

        doc = odmlfmt.Document.create(**doc_attrs)
        for sec in doc_secs:
            doc.append(sec)

        return doc

    def parse_sections(self, section_list):
        odml_sections = []

        for section in section_list:
            sec_attrs = {}
            children_secs = []
            sec_props = []

            for i in section:
                attr = self.is_valid_attribute(i, odmlfmt.Section)
                if attr == 'properties':
                    sec_props = self.parse_properties(section['properties'])
                elif attr == 'sections':
                    children_secs = self.parse_sections(section['sections'])
                elif attr:
                    # Make sure to always use the correct odml format attribute name
                    sec_attrs[odmlfmt.Section.map(attr)] = section[attr]

            sec = odmlfmt.Section.create(**sec_attrs)
            for prop in sec_props:
                sec.append(prop)

            for child_sec in children_secs:
                sec.append(child_sec)

            odml_sections.append(sec)

        return odml_sections

    def parse_properties(self, props_list):
        odml_props = []

        for _property in props_list:
            prop_attrs = {}

            for i in _property:
                attr = self.is_valid_attribute(i, odmlfmt.Property)
                if attr:
                    # Make sure to always use the correct odml format attribute name
                    prop_attrs[odmlfmt.Property.map(attr)] = _property[attr]

            prop = odmlfmt.Property.create(**prop_attrs)
            odml_props.append(prop)

        return odml_props



import unittest
from . import test_samplefile as samplefile


class TestLinks(unittest.TestCase):

    def setUp(self):
        self.doc = samplefile.SampleFileCreator().create_document()
        # for s in self.doc: xmlparser.dumpSection(s)

    def test_link_creation(self):
        obj = self.doc.sections[0].sections[0]
        dst = self.doc.sections[1].sections[1]

        self.assertNotEqual(obj, dst)
        obj.link = "/sec 1/sec 1,1"
        self.assertIsNot(obj, dst)
        self.assertEqual(obj.sections, dst.sections)
        self.assertEqual(obj.properties, dst.properties)

        obj.clean()
        self.assertNotEqual(obj, dst)

    def no_test_circles(self):
        # TODO this currently just works, although, maybe it shouldn't?

        # we cannot allow self-referencing links
        obj = self.doc.sections[0].sections[0]
        dst = self.doc.sections[0]

        samplefile.dumper.dumpSection(dst)

        obj.link = "/sec 0"
        # self.assertEqual(obj.sections, dst.sections) # this will FAIL
        # self.assertEqual(obj.properties, dst.properties)
        obj.clean()

        samplefile.dumper.dumpSection(dst)

    def test_merge(self):
        obj = self.doc.sections[0].sections[0]  # must be an empty section
        dst = self.doc.sections[1]  # .sections[1]
        org = obj.clone()

        obj.link = '/sec 1'
        self.assertEqual(obj.sections, dst.sections)
        self.assertEqual(obj.properties, dst.properties)
        self.assertEqual(obj._merged, dst)

        obj.clean()
        self.assertIsNone(obj._merged)
        self.assertEqual(obj.sections, org.sections)
        self.assertEqual(obj.properties, org.properties)



import os
import unittest

from odml.tools import xmlparser
from odml.tools.parser_utils import ParserException, InvalidVersionException


class TestXMLParser(unittest.TestCase):

    def setUp(self):
        dir_path = os.path.dirname(os.path.realpath(__file__))
        self.basepath = os.path.join(dir_path, "resources")

        self.xml_reader = xmlparser.XMLReader()

    def test_invalid_root(self):
        filename = "invalid_root.xml"
        message = "Expecting <odML>"

        with self.assertRaises(ParserException) as exc:
            _ = self.xml_reader.from_file(os.path.join(self.basepath, filename))

        self.assertIn(message, str(exc.exception))

    def test_missing_version(self):
        filename = "missing_version.xml"
        message = "Could not find format version attribute"

        with self.assertRaises(ParserException) as exc:
            _ = self.xml_reader.from_file(os.path.join(self.basepath, filename))

        self.assertIn(message, str(exc.exception))

    def test_invalid_version(self):
        filename = "invalid_version.xml"

        with self.assertRaises(InvalidVersionException):
            _ = self.xml_reader.from_file(os.path.join(self.basepath, filename))
